# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Standard TC filename formatter."""

# Python Standard Libraries
import logging

from os.path import join as pathjoin, splitext as pathsplitext
from os.path import (
    dirname as pathdirname,
    basename as pathbasename,
)
from datetime import datetime, timedelta
from glob import glob
from os import unlink as osunlink

from geoips.filenames.base_paths import PATHS as gpaths
from geoips.data_manipulations.merge import minrange

LOG = logging.getLogger(__name__)

interface = "filename_formatters"
family = "standard"
name = "tc_fname"


def call(
    area_def,
    xarray_obj,
    product_name,
    coverage=None,
    output_type="png",
    output_type_dir=None,
    product_dir=None,
    product_subdir=None,
    source_dir=None,
    basedir=gpaths["TCWWW"],
    extra_field=None,
    output_dict=None,
):
    """Create standard TC filenames.

    See Also
    --------
    geoips.plugins.modules.filename_formatters.tc_fname.assemble_tc_fname
        This uses the shared utility "assemble_tc_fname", such that a common
        filename can be used by related filename formatters.
    """
    from geoips.sector_utils.utils import is_sector_type

    if area_def and not is_sector_type(area_def, "tc"):
        LOG.warning("NOT a TC sector, skipping TC output")
        return None

    if not product_dir:
        product_dir = product_name

    if not output_type_dir:
        output_type_dir = output_type

    # This allows you to explicitly set matplotlib parameters (colorbars, titles, etc).
    # Overrides were placed in geoimgbase.py to allow using explicitly set values
    # rather than geoimgbase determined defaults.
    # Return reused parameters (min/max vals for normalization, colormaps,
    # matplotlib Normalization)
    # from geoips.xarray_utils.time import get_min_from_xarray_time
    # start_dt = get_min_from_xarray_time(xarray_obj, 'time')
    start_dt = xarray_obj.start_datetime

    if area_def.sector_info["vmax"]:
        intensity = "{0:0.0f}kts".format(area_def.sector_info["vmax"])
    else:
        # This is pulling intensity directly from the deck file,
        # and sometimes it is not defined - if empty, just
        # use "unknown" for intensity
        intensity = "unknown"

    from geoips.plugins.modules.filename_formatters.utils.tc_file_naming import (
        update_extra_field,
    )

    extra = update_extra_field(
        output_dict,
        xarray_obj,
        area_def,
        product_name,
        extra_field_delimiter="-",
        existing_extra_field=extra_field,
        extra_field_resolution=True,
        extra_field_coverage_func=True,
        extra_field_provider=False,
        include_filename_extra_fields=True,
    )

    web_fname = assemble_tc_fname(
        basedir=basedir,
        tc_year=int(area_def.sector_info["storm_year"]),
        tc_basin=area_def.sector_info["storm_basin"],
        tc_stormnum=int(area_def.sector_info["storm_num"]),
        output_type=output_type,
        product_name=product_name,
        product_dir=product_dir,
        product_subdir=product_subdir,
        source_name=xarray_obj.source_name,
        platform_name=xarray_obj.platform_name,
        coverage=coverage,
        product_datetime=start_dt,
        intensity=intensity,
        extra=extra,
        output_type_dir=output_type_dir,
        output_dict=output_dict,
        sector_info=area_def.sector_info,
    )

    return web_fname


def tc_fname_remove_duplicates(fname, mins_to_remove=3, remove_files=False):
    """Remove tc_fname duplicate files.

    Matches storm name, sensor name, platform name, product name, and
    resolution for all files within "mins_to_remove" minutes of the
    current file.  All other fields are wild carded during the matching
    process.
    """
    # 20201010_222325_WP162020_gmi_GPM_89H_40kts_14p16_1p0.png
    # 20201010_222325_WP162020_gmi_GPM_89H_40kts_14p16_1p0.png.yaml
    matching_fnames = []
    removed_fnames = []
    saved_fnames = []
    ext1 = pathsplitext(fname)[-1]
    ext2 = pathsplitext(pathsplitext(fname)[0])[-1]
    # ext3 = pathsplitext(pathsplitext(pathsplitext(fname)[0])[0])[-1]
    if (ext1 == ".png") or (ext1 == ".yaml" and ext2 == ".png"):
        LOG.info(
            "MATCHES EXT FORMAT. png or png.yaml. "
            "Attempting to remove old_tcweb duplicates"
        )
    else:
        LOG.info("NOT REMOVING DUPLICATES. Not tc_web filename, not png or png.yaml.")
        return [], []
    dirname = pathdirname(fname)
    basename = pathbasename(fname)
    parts = basename.split("_")
    if len(parts) != 9:
        LOG.info(
            "NOT REMOVING DUPLICATES. Not tc_web filename, does not contain 9 fields."
        )
        return [], []

    try:
        # 20201010_222325_WP162020_gmi_GPM_89H_40kts_14p16_1p0.png
        yyyymmdd = parts[0]
        hhmnss = parts[1]
        stormname = parts[2]
        sensor = parts[3]
        platform = parts[4]
        product = parts[5]
        intensity = parts[6]
        coverage = parts[7]
        res = parts[8]
        if "p" not in coverage or "p" not in res:
            LOG.info(
                "NOT REMOVING DUPLICATES. Not tc_web filename, "
                "coverage or res not 'NNpNN'."
            )
            return [], []
        if "kts" not in intensity:
            LOG.info(
                "NOT REMOVING DUPLICATES. Not tc_web filename, "
                "intensity does not contain 'kts'."
            )
            return [], []
    except IndexError:
        LOG.info(
            "NOT REMOVING DUPLICATES. Unmatched filename format, "
            "incorrect number of _ delimited fields"
        )
        return [], []
    try:
        fname_dt = datetime.strptime(yyyymmdd + hhmnss, "%Y%m%d%H%M%S")
    except ValueError:
        LOG.info(
            "NOT REMOVING DUPLICATES. Unmatched filename format, "
            "incorrect date time string."
        )
        return [], []
    timediff = timedelta(minutes=mins_to_remove)
    for currdt in minrange(fname_dt - timediff, fname_dt + timediff):
        # 20201010_222325_WP162020_gmi_GPM_89H_40kts_14p16_1p0.png
        dtstr = currdt.strftime(
            "{0}/%Y%m%d_%H%M*_{1}_{2}_{3}_{4}_*_*_{5}".format(
                dirname, stormname, sensor, platform, product, res
            )
        )
        # LOG.info(dtstr)
        matching_fnames += glob(dtstr)
    max_coverage = 0
    min_dt = None
    for matching_fname in matching_fnames:
        # 20201010_222325_WP162020_gmi_GPM_89H_40kts_14p16_1p0.png
        parts = pathbasename(matching_fname).split("_")
        coverage = float(parts[7].replace("p", "."))
        start_dt = datetime.strptime(parts[0] + parts[1][0:6], "%Y%m%d%H%M%S")
        max_coverage = max(coverage, max_coverage)
        if min_dt is None:
            min_dt = start_dt
        else:
            min_dt = min(start_dt, min_dt)

    gotone = False
    LOG.info("CHECKING DUPLICATE FILES")
    for matching_fname in list(set(matching_fnames)):
        # 20201010_222325_WP162020_gmi_GPM_89H_40kts_14p16_1p0.png
        parts = pathbasename(matching_fname).split("_")
        coverage = float(parts[7].replace("p", "."))
        start_dt = datetime.strptime(parts[0] + parts[1][0:6], "%Y%m%d%H%M%S")
        # Priority to delete lower coverage
        if coverage < max_coverage:
            removed_fnames += [matching_fname]
            # Test it out for a bit first
            if remove_files is True:
                LOG.info(
                    "DELETING DUPLICATE FILE with less coverage %s < %s %s",
                    coverage,
                    max_coverage,
                    matching_fname,
                )
                try:
                    osunlink(matching_fname)
                except FileNotFoundError as resp:
                    LOG.warning(
                        "FAILDELETE %s: File %s did not exist, "
                        "someone must have deleted it for us?",
                        matching_fname,
                        str(resp),
                    )

            else:
                LOG.info(
                    "TEST DELETING DUPLICATE FILE with less coverage %s < %s %s",
                    coverage,
                    max_coverage,
                    matching_fname,
                )
        elif start_dt > min_dt:
            removed_fnames += [matching_fname]
            # Test it out for a bit first
            if remove_files is True:
                LOG.info(
                    "DELETING DUPLICATE FILE with later start_dt %s > %s %s",
                    start_dt,
                    min_dt,
                    matching_fname,
                )
                try:
                    osunlink(matching_fname)
                except FileNotFoundError as resp:
                    LOG.warning(
                        "FAILDELETE %s: File %s did not exist, "
                        "someone must have deleted it for us?",
                        matching_fname,
                        str(resp),
                    )

            else:
                LOG.info(
                    "TEST DELETING DUPLICATE FILE with later start_dt %s > %s %s",
                    start_dt,
                    min_dt,
                    matching_fname,
                )
        elif gotone is True:
            removed_fnames += [matching_fname]
            # Test it out for a bit first
            if remove_files is True:
                LOG.info(
                    "DELETING DUPLICATE FILE with same coverage %s = %s "
                    "and same start_dt %s = %s, %s",
                    coverage,
                    max_coverage,
                    start_dt,
                    min_dt,
                    matching_fname,
                )
                try:
                    osunlink(matching_fname)
                except FileNotFoundError as resp:
                    LOG.warning(
                        "FAILDELETE %s: File %s did not exist, "
                        "someone must have deleted it for us?",
                        matching_fname,
                        str(resp),
                    )

            else:
                LOG.info(
                    "TEST DELETING DUPLICATE FILE with same coverage %s = %s "
                    "and same start_dt %s = %s, %s",
                    coverage,
                    max_coverage,
                    start_dt,
                    min_dt,
                    matching_fname,
                )
        else:
            if len(matching_fnames) == 1:
                LOG.info(
                    "SAVING DUPLICATE FILE (only one!) with max coverage %s %s",
                    max_coverage,
                    matching_fname,
                )
            else:
                LOG.info(
                    "SAVING DUPLICATE FILE with max coverage %s %s",
                    max_coverage,
                    matching_fname,
                )
            saved_fnames += [matching_fname]
            gotone = True
        # from IPython import embed as shell; shell()

    return removed_fnames, saved_fnames


def assemble_tc_fname(
    basedir,
    tc_year,
    tc_basin,
    tc_stormnum,
    output_type,
    product_name,
    source_name,
    platform_name,
    coverage,
    product_datetime,
    intensity=None,
    extra=None,
    output_type_dir=None,
    product_dir=None,
    product_subdir=None,
    output_dict=None,
    sector_info=None,
):
    """Produce full output product path from product / sensor specifications.

    tc web paths are of the format:
        <basedir>/tc<tc_year>/<tc_basin>/<tc_basin><tc_stormnum><tc_year>/
          <output_type>/<product_name>/<platform_name>/
    tc web filenames are of the format:
        <date{%Y%m%d%H%M>_<tc_basin><tc_stormnum><tc_year>_<source_name>_
          <platform_name>_<product_name>_<intensity>_<coverage>_
          <extra>.<output_type>

    Parameters
    ----------
    basedir : str
        Base directory for output file.
    tc_year : int
        Full 4 digit storm year
    tc_basin : str
        2 character basin designation
          SH Southern Hemisphere
          WP West Pacific
          EP East Pacific
          CP Central Pacific
          IO Indian Ocean
          AL Atlantic
    tc_stormnum : int
        2 digit storm number
          90 through 99 for invests
          01 through 69 for named storms
    output_type : str
        file extension type
    product_name : str
        Name of product
    source_name : str
        Name of data source (sensor)
    platform_name : str
        Name of platform (satellite)
    coverage : float
        Image coverage, float between 0.0 and 100.0
    product_datetime : datetime.datetime
        Datetime object - start time of data used to generate product
    output_type_dir : str
        Default output_type, dir name
    product_dir : str
        Default product_name, dir name

    Returns
    -------
    str
        full path to output file
    """
    if not output_type_dir:
        output_type_dir = output_type
    if not product_dir:
        product_dir = product_name
    if not product_subdir:
        product_subdir = platform_name

    from geoips.plugins.modules.filename_formatters.utils.tc_file_naming import (
        tc_storm_basedir,
    )

    path = pathjoin(
        tc_storm_basedir(
            basedir,
            tc_year,
            tc_basin,
            tc_stormnum,
            output_dict=output_dict,
            sector_info=sector_info,
        ),
        output_type_dir,
        product_dir,
        product_subdir,
    )
    fname = "_".join(
        [
            product_datetime.strftime("%Y%m%d"),
            product_datetime.strftime("%H%M%S"),
            "{0}{1:02d}{2:04d}".format(tc_basin, tc_stormnum, tc_year),
            source_name,
            platform_name,
            product_name,
            str(intensity),
            "{0:0.2f}".format(coverage).replace(".", "p"),
            str(extra),
        ]
    )
    fname = "{0}.{1}".format(fname, output_type)
    return pathjoin(path, fname)



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Surface Winds plotted as Barbs in Knots.

Data manipulation steps for surface winds products.
This algorithm expects surface wind speeds in units of kts
"""
import logging

# install libraries
import numpy as np

LOG = logging.getLogger(__name__)

interface = "algorithms"
family = "list_numpy_to_numpy"
name = "windbarbs"


def call(
    arrays,
    output_data_range=None,
    input_units=None,
    output_units=None,
    min_outbounds="crop",
    max_outbounds="crop",
    norm=False,
    inverse=False,
):
    """Windbarbs product algorithm data manipulation steps.

    This algorithm expects input windspeed with units "kts" and returns in "kts"

    Parameters
    ----------
    arrays : list of numpy.ndarray
        * list of numpy.ndarray or numpy.MaskedArray of channel data,
            in order of sensor "channels" list
        * kts
    output_data_range : list of float, default=None
        * list of min and max value for wind speeds (kts)
        * defaults to None, which results in using data.min and data.max.
    input_units : str, default=None
        * Units of input data, for applying necessary conversions
        * defaults to None, resulting in no unit conversions.
    output_units : str, default=None
        * Units of output data, for applying necessary conversions
        * defaults to None, resulting in no unit conversions.
    min_outbounds : str, default='crop'
        * Method to use when applying bounds.  Valid values are:

            * retain: keep all pixels as is
            * mask: mask all pixels that are out of range
            * crop: set all out of range values to either min_val or max_val
              as appropriate
    max_outbounds : str, default='crop'
        * Method to use when applying bounds.  Valid values are:

            * retain: keep all pixels as is
            * mask: mask all pixels that are out of range
            * crop: set all out of range values to either min_val or max_val
              as appropriate
    norm : bool, default=False
        * Boolean flag indicating whether to normalize (True) or not (False)

            * If True, returned data will be in the range from 0 to 1
            * If False, returned data will be in the range from min_val to
              max_val
    inverse : bool, default=False
        * Boolean flag indicating whether to inverse (True) or not (False)

            * If True, returned data will be inverted
            * If False, returned data will not be inverted

    Returns
    -------
    numpy.ndarray
        numpy.ndarray or numpy.MaskedArray of appropriately scaled channel data,
        dstacked as follows:

        * (spd, direction, rain_flag)
        * spd in kts
        * direction in degrees
        * rain_flag 0 or 1
    """
    spd = arrays[0]
    if output_data_range is None:
        output_data_range = (spd.min(), spd.max)
    direction = arrays[1]
    if len(arrays) > 2:
        rain_flag = arrays[2]
    else:
        rain_flag = np.zeros(arrays[1].shape)

    from geoips.data_manipulations.conversions import unit_conversion

    spd = unit_conversion(spd, input_units, output_units)

    from geoips.data_manipulations.corrections import apply_data_range

    spd = apply_data_range(
        spd,
        min_val=output_data_range[0],
        max_val=output_data_range[1],
        min_outbounds=min_outbounds,
        max_outbounds=max_outbounds,
        norm=norm,
        inverse=inverse,
    )
    return np.ma.dstack((spd, direction, rain_flag)).squeeze()



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Advanced Himawari Imager Data Reader."""

# Python Standard Libraries
from datetime import datetime, timedelta
from glob import glob
import logging
import os
import re
from struct import unpack

# Third-Party Libraries
import numpy as np
import xarray
from scipy.ndimage import zoom

# GeoIPS Libraries
from geoips.interfaces import readers
from geoips.utils.memusg import print_mem_usage
from geoips.utils.context_managers import import_optional_dependencies
from geoips.plugins.modules.readers.utils.geostationary_geolocation import (
    get_geolocation_cache_filename,
    get_geolocation,
)

LOG = logging.getLogger(__name__)

with import_optional_dependencies(loglevel="info"):
    """Attempt to import a package and print to LOG.info if the import fails."""
    import numexpr as ne

try:
    nprocs = 6
    ne.set_num_threads(nprocs)
except Exception:
    LOG.info(
        "Failed numexpr.set_num_threads in %s. If numexpr is not installed and you"
        " need it, install it.",
        __file__,
    )

DONT_AUTOGEN_GEOLOCATION = False
if os.getenv("DONT_AUTOGEN_GEOLOCATION"):
    DONT_AUTOGEN_GEOLOCATION = True

# These should be added to the data file object
BADVALS = {
    "Off_Of_Disk": -999.9,
    "Error": -999.8,
    "Out_Of_Valid_Range": -999.7,
    "Root_Test": -999.6,
    "Unitialized": -9999.9,
}

DATASET_INFO = {
    "LOW": [
        "B05",
        "B06",
        "B07",
        "B08",
        "B09",
        "B10",
        "B11",
        "B12",
        "B13",
        "B14",
        "B15",
        "B16",
    ],
    "MED": ["B01", "B02", "B04"],
    "HIGH": ["B03"],
}
ROIS = {"LOW": 3000, "MED": 2000, "HIGH": 1000}
SAMPLE_DISTANCE_KM = {"LOW": 2, "MED": 1, "HIGH": 0.5}
ALL_CHANS = {
    "LOW": [
        "B05Rad",
        "B05Ref",  # 1.61um
        "B06Rad",
        "B06Ref",  # 2.267um
        "B07Rad",
        "B07BT",  # 3.8853um
        "B08Rad",
        "B08BT",  # 6.2429um
        "B09Rad",
        "B09BT",  # 6.9410um
        "B10Rad",
        "B10BT",  # 7.3467um
        "B11Rad",
        "B11BT",  # 8.5926um
        "B12Rad",
        "B12BT",  # 9.6372um
        "B13Rad",
        "B13BT",  # 10.4073um
        "B14Rad",
        "B14BT",  # 11.2395um
        "B15Rad",
        "B15BT",  # 12.3806um
        "B16Rad",
        "B16BT",
    ],  # 13.2807um
    "MED": [
        "B01Rad",
        "B01Ref",  # 0.47063um
        "B02Rad",
        "B02Ref",  # 0.51000um
        "B04Rad",
        "B04Ref",
    ],  # 0.85670um
    "HIGH": ["B03Rad", "B03Ref"],  # 0.63914um
}

ALL_GVARS = {
    "LOW": [
        "latitude",
        "longitude",
        "solar_zenith_angle",
        "solar_azimuth_angle",
        "satellite_zenith_angle",
        "satellite_azimuth_angle",
    ],
    "MED": [
        "latitude",
        "longitude",
        "solar_zenith_angle",
        "solar_azimuth_angle",
        "satellite_zenith_angle",
        "satellite_azimuth_angle",
    ],
    "HIGH": [
        "latitude",
        "longitude",
        "solar_zenith_angle",
        "solar_azimuth_angle",
        "satellite_zenith_angle",
        "satellite_azimuth_angle",
    ],
}

interface = "readers"
family = "standard"
name = "ahi_hsd"
source_names = ["ahi"]


class AutoGenError(Exception):
    """Raise exception on geolocation autogeneration error."""

    pass


def findDiff(d1, d2, path=""):
    """Find diff."""
    for k in d1.keys():
        if k not in d2:
            LOG.info(path, ":")
            LOG.info(k + " as key not in d2", "\n")
        else:
            if type(d1[k]) is dict:
                if path == "":
                    path = k
                else:
                    path = path + "->" + k
                findDiff(d1[k], d2[k], path)
            else:
                if np.all(d1[k] != d2[k]):
                    LOG.info(path, ":")
                    LOG.info(" - ", k, " : ", d1[k])
                    LOG.info(" + ", k, " : ", d2[k])


def metadata_to_datetime(metadata, time_var="ob_start_time"):
    """Use information from block_01 to get the image datetime."""
    ost = metadata["block_01"][time_var]
    otl = metadata["block_01"]["ob_timeline"]
    dt = datetime(1858, 11, 17, 00, 00, 00)
    dt += timedelta(days=np.floor(ost))
    dt += timedelta(hours=int(otl // 100), minutes=int(otl % 100))
    return dt


def _get_geolocation_metadata(metadata):
    """
    Gather all of the metadata used in creating geolocation data for input file.

    This is split out so we can easily create a hash of the data for creation
    of a unique filename. This allows us to avoid recalculation of angles that
    have already been calculated.
    """
    geomet = {}
    # Used for cached file paths
    geomet["platform_name"] = metadata["block_01"]["satellite_name"].lower()
    geomet["ob_area"] = metadata["block_01"]["ob_area"]
    # Used for cached filenames
    geomet["scene"] = geomet["ob_area"]
    geomet["num_lines"] = int(
        metadata["block_02"]["num_lines"] * metadata["block_07"]["num_segments"]
    )
    geomet["num_samples"] = int(metadata["block_02"]["num_samples"])
    geomet["lfac"] = float(metadata["block_03"]["LFAC"])
    geomet["loff"] = float(metadata["block_03"]["LOFF"])
    geomet["cfac"] = float(metadata["block_03"]["CFAC"])
    geomet["res_km"] = np.ceil(2.0 * 40000000.0 / geomet["cfac"]) / 2.0
    geomet["roi_factor"] = 5  # roi = res * roi_factor, was 10
    geomet["coff"] = float(metadata["block_03"]["COFF"])
    geomet["Rs"] = float(metadata["block_03"]["earth_to_sat_radius"])
    geomet["H_m"] = float(geomet["Rs"] * 1000.0)
    geomet["Sd_coeff"] = float(metadata["block_03"]["Sd_coeff"])
    geomet["ecc"] = float(metadata["block_03"]["r3"])
    geomet["sub_lon"] = float(metadata["block_03"]["sub_lon"])
    geomet["lon0"] = geomet["sub_lon"]
    return geomet


def get_latitude_longitude(metadata, BADVALS, area_def=None):
    """
    Get latitudes and longitudes.

    This routine accepts a dictionary containing metadata as read from an HSD
    format file and returns latitudes and longitudes for a full disk image.

    Note: This code has been adapted from Dan Lindsey's Fortran90 code.
    This was done in three steps that ultimately culminated in faster, but more
    difficult to understand code.  If you plan to edit this, I recommend that
    you return to Dan's original code, then explore the commented code here,
    then finally, look at the single-command statements that are currently being
    used.
    """
    # If the filename format needs to change for the pre-generated geolocation
    # files, please discuss prior to changing.  It will force recreation of all
    # files, which can be problematic for large numbers of sectors
    fname = get_geolocation_cache_filename("GEOLL", metadata)
    if not os.path.isfile(fname):
        if (
            area_def is not None
            and DONT_AUTOGEN_GEOLOCATION
            and "tc2019" not in area_def.area_id
        ):
            msg = (
                "GETGEO Requested NO AUTOGEN GEOLOCATION. "
                + "Could not create latlonfile for ad {}: {}"
            ).format(metadata["ob_area"], fname)
            LOG.error(msg)
            raise AutoGenError(msg)

        LOG.debug("Calculating latitudes and longitudes.")

        sclunit = 1.525878906250000e-05  # NOQA

        # Constants
        LOG.info("    LATLONCALC Building constants.")
        pi = np.pi
        rad2deg = 180.0 / pi  # NOQA
        deg2rad = pi / 180.0  # NOQA
        num_lines = metadata["num_lines"]
        num_samples = metadata["num_samples"]
        lfac = metadata["lfac"]  # NOQA
        loff = metadata["loff"]  # NOQA
        cfac = metadata["cfac"]  # NOQA
        coff = metadata["coff"]  # NOQA
        Rs = metadata["Rs"]  # NOQA
        Sd_coeff = metadata["Sd_coeff"]  # NOQA
        ecc = metadata["ecc"]  # NOQA
        sub_lon = metadata["sub_lon"]  # NOQA

        # first_line = df.metadata['block_07']['segment_first_line'][0]
        first_line = 0
        # last_line = first_line + num_lines
        last_line = num_lines
        line_step = 1

        first_sample = 0
        last_sample = num_samples
        sample_step = 1

        # Create cartesian grid
        LOG.info("    LATLONCALC Creating cartesian grid")
        x, y = np.meshgrid(
            np.arange(first_sample, last_sample, sample_step),
            np.arange(first_line, last_line, line_step),
        )
        # Changing to use numexpr rather than numpy.  Appears to speed up each statement
        # by about five times.
        #
        # In [8]: timeit -n100 deg2rad *
        # (np.array(x, dtype=np.float) - coff)/(sclunit * cfac)
        # 100 loops, best of 3: 96.6 ms per loop
        #
        # In [9]: timeit -n100 ne.evaluate('deg2rad*(x-coff)/(sclunit*cfac)')
        # 100 loops, best of 3: 20 ms per loop
        x = ne.evaluate("deg2rad * (x - coff) / (sclunit * cfac)")  # NOQA
        y = ne.evaluate("deg2rad*(y - loff)/(sclunit * lfac)")  # NOQA
        # # Improvement here is from 132ms for np.sin(x) to 23.5ms for
        # ne.evaluate('sin(x)')
        LOG.info("    LATLONCALC Calculating sines and cosines")
        # sin_x = ne.evaluate('sin(x)')  # NOQA
        # sin_y = ne.evaluate('sin(y)')  # NOQA
        # cos_x = ne.evaluate('cos(x)')  # NOQA
        # cos_y = ne.evaluate('cos(y)')  # NOQA

        # Calculate surface distance (I think)
        # Improvement here is from 200ms for numpy to 16.9ms for ne
        LOG.debug("Calculating Sd")
        Sd = ne.evaluate(
            "(Rs * cos(x) * cos(y))**2 - (cos(y)**2 + ecc * sin(y)**2) * Sd_coeff"
        )
        # No real savings on these lines.  Leave alone.
        Sd[Sd < 0.0] = 0.0
        Sd **= 0.5

        # # Good data mask
        # good = Sd != 0

        # # Now I'm lost, but it seems to work.  Comes from the Himawari-8 users's guide
        # # Original version with excess calculations
        LOG.debug("Calculating Sn")
        # Sn = ne.evaluate('(Rs * cos_x * cos_y-Sd)/(cos_y**2 + ecc * sin_y**2)')
        # # NOQA
        LOG.debug("Calculating S1")
        # S1 = ne.evaluate('Rs - (Sn * cos_x * cos_y)')  # NOQA
        LOG.debug("Calculating S2")
        # S2 = ne.evaluate('Sn * sin_x * cos_y')  # NOQA
        LOG.debug("Calculating S3")
        # S3 = ne.evaluate('-Sn * sin_y')  # NOQA
        LOG.debug("Calculating Sxy")
        # Sxy = ne.evaluate('(S1**2 + S2**2)**0.5')  # NOQA

        # # if hasattr(self, '_temp_latitudes_arr'):
        # #     lats = self._temp_latitudes_arr
        # # else:
        LOG.debug("    LATLONCALC Allocating latitudes")
        # lats = np.full((num_lines, num_samples), df.BADVALS['Off_Of_Disk'])
        # # if hasattr(self, '_temp_longitudes_arr'):
        # #     lons = self._temp_longitudes_arr
        # # else:
        LOG.debug("    LATLONCALC Allocating longitudes")
        # lons = np.full((num_lines, num_samples), df.BADVALS['Off_Of_Disk'])

        # # It may help to figure out how to index into an array in numeval
        # # Improves from 663ms for numpy to 329ms for numeval
        # LOG.debug('Calculating latitudes')
        # lats[good] = ne.evaluate('rad2deg*arctan(ecc*S3/Sxy)')[good]
        # # Improves from 669ms for numpy to 301ms for numeval
        # LOG.debug('Calculating longitudes')
        # lons[good] = ne.evaluate('rad2deg*arctan(S2/S1)+sub_lon')[good]
        # # No real savings on these lines.  Leave alone.
        # lons[lons > 180.0] -= 360
        # # self._latitudes = lats
        # # self._longitudes = lons

        # The following equations have been combined from above.
        # The more we can fit into a single equation, the faster things will be.
        # I know this makes things ugly, but hopefully this will never have to be
        # edited.
        LOG.info("    LATLONCALC Calculating latitudes")
        bad = Sd == 0
        lats = ne.evaluate(
            "rad2deg*arctan(-ecc*(Rs*cos(x)*cos(y)-Sd)"
            "/ (cos(y)**2+ecc*sin(y)**2) * sin(y)"
            "/ ((Rs-(Rs*cos(x)*cos(y)-Sd)/(cos(y)**2+ecc*sin(y)**2)*cos(x)*cos(y))**2"
            "+ ((Rs*cos(x)*cos(y)-Sd)"
            "/(cos(y)**2+ecc*sin(y)**2)*sin(x)*cos(y))**2)**0.5)"
        )
        lats[bad] = BADVALS["Off_Of_Disk"]
        LOG.info("    LATLONCALC Calculating longitudes")
        lons = ne.evaluate(
            "rad2deg*arctan(((Rs*cos(x)*cos(y)-Sd)"
            "/ (cos(y)**2 + ecc*sin(y)**2))*sin(x)*cos(y)"
            "/ (Rs-((Rs*cos(x)*cos(y)-Sd)"
            "/ (cos(y)**2 + ecc*sin(y)**2))*cos(x)*cos(y))) + sub_lon"
        )
        lons[bad] = BADVALS["Off_Of_Disk"]
        lons[lons > 180.0] -= 360
        LOG.debug("Done calculating latitudes and longitudes")

        with open(fname, "w") as df:
            lats.tofile(df)
            lons.tofile(df)
        # Switch to xarray based geolocation files
        # ds = xarray.Dataset({'latitude':(['x','y'],lats),
        #                      'longitude':(['x','y'],lons)})
        # ds.to_netcdf(fname)

    # Create a memmap to the lat/lon file
    # Nothing will be read until explicitly requested
    # We are mapping this here so that the lats and lons are available when
    # calculating satlelite angles
    LOG.info(
        "GETGEO memmap to {} : lat/lon file for {}".format(fname, metadata["ob_area"])
    )

    shape = (metadata["num_lines"], metadata["num_samples"])
    offset = 8 * metadata["num_samples"] * metadata["num_lines"]
    lats = np.memmap(fname, mode="r", dtype=np.float64, offset=0, shape=shape)
    lons = np.memmap(fname, mode="r", dtype=np.float64, offset=offset, shape=shape)
    # Switch to xarray based geolocation files
    # saved_xarray = xarray.load_dataset(fname)
    # lons = saved_xarray['longitude'].to_masked_array()
    # lats = saved_xarray['latitude'].to_masked_array()

    return lats, lons


def _get_metadata_block_info(df):
    """
    Get metadata block info.

    Returns a dictionary whose keys represent metadata block numbers and whose
    values are tuples containing the block's starting byte number and the
    block's length in bytes.
    """
    # Initialize the first block's data
    block_info = {1: (0, 0)}
    # print(df.name)

    # Loop over blocks and determine their sizes
    for blockind in range(1, 12):
        # Make sure we are at the start of the block
        block_start = block_info[blockind][0]
        df.seek(block_start)

        # Read the block number and block length
        block_num = unpack("B", df.read(1))[0]
        if block_num != blockind:
            raise IOError(
                "Unexpected block number encountered. "
                f"Expected {blockind}, but got {block_num}. File {df.name}"
            )
        block_length = unpack("H", df.read(2))[0]

        # Add block_length for the current block
        block_info[blockind] = (block_start, block_length)

        # Add entry for the NEXT block
        if blockind < 11:
            block_info[blockind + 1] = (block_start + block_length, 0)

    return block_info


def _get_metadata_block_01(df, block_info):
    """
    Get metadata block 01.

    Parse the first metadata block from an AHI data file and return a dictionary
    containing that metadata.
    """
    # Get the block start and length in bytes
    block_start, block_length = block_info[1]

    # Seek to the start of the block and read the block in its entirety
    df.seek(block_start)
    data = df.read(block_length)

    # Create dictionary for this block
    block_data = {}
    block_data["block_name"] = "basic_information"
    block_data["block_num"] = np.fromstring(data[0:1], dtype="uint8")[0]
    block_data["block_length"] = np.fromstring(data[1:3], dtype="uint16")[0]
    block_data["num_headers"] = np.fromstring(data[3:5], dtype="uint16")[0]
    block_data["byte_order"] = np.fromstring(data[5:6], dtype="uint8")[0]
    block_data["satellite_name"] = data[6:22].decode("ascii").replace("\x00", "")
    block_data["processing_center"] = data[22:38].decode("ascii").replace("\x00", "")
    block_data["ob_area"] = data[38:42].decode("ascii").replace("\x00", "")
    block_data["other_ob_info"] = data[42:44].decode("ascii").replace("\x00", "")
    block_data["ob_timeline"] = np.fromstring(data[44:46], dtype="uint16")[0]
    block_data["ob_start_time"] = np.fromstring(data[46:54], dtype="float64")[0]
    block_data["ob_end_time"] = np.fromstring(data[54:62], dtype="float64")[0]
    block_data["creation_time"] = np.fromstring(data[62:70], dtype="float64")[0]
    block_data["total_header_length"] = np.fromstring(data[70:74], dtype="uint32")[0]
    block_data["total_data_length"] = np.fromstring(data[74:78], dtype="uint32")[0]
    block_data["quality_flag_1"] = np.fromstring(data[78:79], dtype="uint8")[0]
    block_data["quality_flag_2"] = np.fromstring(data[79:80], dtype="uint8")[0]
    block_data["quality_flag_3"] = np.fromstring(data[80:81], dtype="uint8")[0]
    block_data["quality_flag_4"] = np.fromstring(data[81:82], dtype="uint8")[0]
    block_data["file_format_version"] = data[82:114].decode("ascii").replace("\x00", "")
    block_data["file_name"] = data[114:242].decode("ascii").replace("\x00", "")
    block_data["spare"] = data[242:282].decode("ascii").replace("\x00", "")

    return block_data


def _get_metadata_block_02(df, block_info):
    """
    Get metadata block 02.

    Parse the second metadata block from an AHI data file and return a
    dictionary containing that metadata.
    """
    # Get the block start and length in bytes
    block_start, block_length = block_info[2]

    # Seek to the start of the block and read the block in its entirety
    df.seek(block_start)
    data = df.read(block_length)

    # Create dictionary for this block
    block_data = {}
    block_data["block_name"] = "data_information"
    block_data["block_num"] = np.fromstring(data[0:1], dtype="uint8")[0]
    block_data["block_length"] = np.fromstring(data[1:3], dtype="uint16")[0]
    block_data["bits_per_pixel"] = np.fromstring(data[3:5], dtype="uint16")[0]
    block_data["num_samples"] = np.fromstring(data[5:7], dtype="uint16")[0]
    block_data["num_lines"] = np.fromstring(data[7:9], dtype="uint16")[0]
    block_data["compression_flag"] = np.fromstring(data[9:10], dtype="uint8")[0]
    block_data["spare"] = data[10:50].decode("ascii").replace("\x00", "")

    return block_data


def _get_metadata_block_03(df, block_info):
    """
    Get metadata block 03.

    Parse the third metadata block from an AHI data file and return a
    dictionary containing that metadata.
    """
    # Get the block start and length in bytes
    block_start, block_length = block_info[3]

    # Seek to the start of the block and read the block in its entirety
    df.seek(block_start)
    data = df.read(block_length)

    # Create dictionary for this block
    block_data = {}
    block_data["block_name"] = "projection_information"
    block_data["block_num"] = np.fromstring(data[0:1], dtype="uint8")[0]
    block_data["block_length"] = np.fromstring(data[1:3], dtype="uint16")[0]
    block_data["sub_lon"] = np.fromstring(data[3:11], dtype="float64")[0]
    # Column sclaing factor
    block_data["CFAC"] = np.fromstring(data[11:15], dtype="uint32")[0]
    # Line scaling factor
    block_data["LFAC"] = np.fromstring(data[15:19], dtype="uint32")[0]
    # Column offset
    block_data["COFF"] = np.fromstring(data[19:23], dtype="float32")[0]
    # Line offset
    block_data["LOFF"] = np.fromstring(data[23:27], dtype="float32")[0]
    # Distance to earth's center
    block_data["earth_to_sat_radius"] = np.fromstring(data[27:35], dtype="float64")[0]
    # Radius of earth at equator
    block_data["equator_radius"] = np.fromstring(data[35:43], dtype="float64")[0]
    # Radius of earth at pole
    block_data["pole_radius"] = np.fromstring(data[43:51], dtype="float64")[0]
    # (R_eq**2 - R_pol**2)/R_eq**2
    block_data["r1"] = np.fromstring(data[51:59], dtype="float64")[0]
    # R_pol**2/R_eq**2
    block_data["r2"] = np.fromstring(data[59:67], dtype="float64")[0]
    # R_eq**2/R_pol**2
    block_data["r3"] = np.fromstring(data[67:75], dtype="float64")[0]
    # Coefficient for S_d(R_s**2 - R_eq**2)
    block_data["Sd_coeff"] = np.fromstring(data[75:83], dtype="float64")[0]
    block_data["resampling_types"] = np.fromstring(data[83:85], dtype="uint16")[0]
    block_data["resampling_size"] = np.fromstring(data[85:87], dtype="uint16")[0]
    block_data["spare"] = np.fromstring(data[87:127], dtype="uint16")

    return block_data


def _get_metadata_block_04(df, block_info):
    """
    Get metadata block 04.

    Parse the 4th metadata block from an AHI data file and return a
    dictionary containing that metadata.
    """
    # Get the block start and length in bytes
    block_start, block_length = block_info[4]

    # Seek to the start of the block and read the block in its entirety
    df.seek(block_start)
    data = df.read(block_length)

    # Create dictionary for this block
    block_data = {}
    block_data["block_name"] = "navigation_information"
    block_data["block_num"] = np.fromstring(data[0:1], dtype="uint8")[0]
    block_data["block_length"] = np.fromstring(data[1:3], dtype="uint16")[0]
    block_data["nav_info_time"] = np.fromstring(data[3:11], dtype="float64")[0]
    block_data["SSP_lon"] = np.fromstring(data[11:19], dtype="float64")[0]
    block_data["SSP_lat"] = np.fromstring(data[19:27], dtype="float64")[0]
    block_data["earthcenter_to_sat_dist"] = np.fromstring(data[27:35], dtype="float64")[
        0
    ]
    block_data["nadir_lon"] = np.fromstring(data[35:43], dtype="float64")[0]
    block_data["nadir_lat"] = np.fromstring(data[43:51], dtype="float64")[0]
    block_data["sun_pos"] = np.fromstring(data[51:75], dtype="float64")
    block_data["moon_pos"] = np.fromstring(data[75:99], dtype="float64")
    block_data["spare"] = np.fromstring(data[99:139], dtype="float64")

    return block_data


def _get_metadata_block_05(df, block_info):
    """
    Get metadata block 05.

    Parse the 5th metadata block from an AHI data file and return a
    dictionary containing that metadata.
    """
    # Get the block start and length in bytes
    block_start, block_length = block_info[5]

    # Seek to the start of the block and read the block in its entirety
    df.seek(block_start)
    data = df.read(block_length)

    # Create dictionary for this block
    block_data = {}
    block_data["block_name"] = "calibration_information"
    block_data["block_num"] = np.fromstring(data[0:1], dtype="uint8")[0]
    block_data["block_length"] = np.fromstring(data[1:3], dtype="uint16")[0]
    block_data["band_number"] = np.fromstring(data[3:5], dtype="uint16")[0]
    block_data["cent_wavelenth"] = np.fromstring(data[5:13], dtype="float64")[0]
    block_data["valid_bits_per_pixel"] = np.fromstring(data[13:15], dtype="uint16")[0]
    block_data["count_badval"] = np.fromstring(data[15:17], dtype="uint16")[0]
    block_data["count_outside_scan"] = np.fromstring(data[17:19], dtype="uint16")[0]
    block_data["gain"] = np.fromstring(data[19:27], dtype="float64")[0]
    block_data["offset"] = np.fromstring(data[27:35], dtype="float64")[0]
    if block_data["band_number"] in range(7, 17):
        block_data["c0"] = np.fromstring(data[35:43], dtype="float64")[0]
        block_data["c1"] = np.fromstring(data[43:51], dtype="float64")[0]
        block_data["c2"] = np.fromstring(data[51:59], dtype="float64")[0]
        block_data["C0"] = np.fromstring(data[59:67], dtype="float64")[0]
        block_data["C1"] = np.fromstring(data[67:75], dtype="float64")[0]
        block_data["C2"] = np.fromstring(data[75:83], dtype="float64")[0]
        block_data["speed_of_light"] = np.fromstring(data[83:91], dtype="float64")[0]
        block_data["planck_const"] = np.fromstring(data[91:99], dtype="float64")[0]
        block_data["boltz_const"] = np.fromstring(data[99:107], dtype="float64")[0]
        # block_data['spare'] = np.fromstring(data[107:147], dtype='float64')
    else:
        block_data["c_prime"] = np.fromstring(data[35:43], dtype="float64")[0]
        block_data["spare"] = np.fromstring(data[43:147], dtype="float64")

    return block_data


def _get_metadata_block_06(df, block_info):
    """
    Get metadata block 06.

    Parse the 6th metadata block from an AHI data file and return a
    dictionary containing that metadata.
    """
    # Get the block start and length in bytes
    block_start, block_length = block_info[6]

    # Seek to the start of the block and read the block in its entirety
    df.seek(block_start)
    data = df.read(block_length)

    # Create dictionary for this block
    block_data = {}
    block_data["block_name"] = "intercalibration_information"
    block_data["block_num"] = np.fromstring(data[0:1], dtype="uint8")[0]
    block_data["block_length"] = np.fromstring(data[1:3], dtype="uint16")[0]
    # Global Space-based Inter-Calibration System (GSICS) calibration coefficients
    block_data["GSICS_intercept"] = np.fromstring(data[3:11], dtype="float64")[0]
    block_data["GSICS_slope"] = np.fromstring(data[11:19], dtype="float64")[0]
    block_data["GSICS_quadratic"] = np.fromstring(data[19:27], dtype="float64")[0]
    block_data["radiance_bias"] = np.fromstring(data[27:35], dtype="float64")[0]
    block_data["bias_uncert"] = np.fromstring(data[35:43], dtype="float64")[0]
    block_data["standard_radiance"] = np.fromstring(data[43:51], dtype="float64")[0]
    block_data["GSICS_valid_start"] = np.fromstring(data[51:59], dtype="float64")[0]
    block_data["GSICS_valid_end"] = np.fromstring(data[59:67], dtype="float64")[0]
    block_data["GSICS_upper_limit"] = np.fromstring(data[67:71], dtype="float32")[0]
    block_data["GSICS_lower_limit"] = np.fromstring(data[71:75], dtype="float32")[0]
    block_data["GSICS_filename"] = data[75:203].decode("ascii").replace("\x00", "")
    block_data["spare"] = data[203:259].decode("ascii").replace("\x00", "")

    return block_data


def _get_metadata_block_07(df, block_info):
    """
    Get metadata block 07.

    Parse the 7th metadata block from an AHI data file and return a
    dictionary containing that metadata.
    """
    # Get the block start and length in bytes
    block_start, block_length = block_info[7]

    # Seek to the start of the block and read the block in its entirety
    df.seek(block_start)
    data = df.read(block_length)

    # Create dictionary for this block
    block_data = {}
    block_data["block_name"] = "segment_information"
    block_data["block_num"] = np.fromstring(data[0:1], dtype="uint8")[0]
    block_data["block_length"] = np.fromstring(data[1:3], dtype="uint16")[0]
    block_data["num_segments"] = np.fromstring(data[3:4], dtype="uint8")[0]
    block_data["segment_number"] = np.fromstring(data[4:5], dtype="uint8")[0]
    block_data["segment_first_line"] = np.fromstring(data[5:7], dtype="uint16")[0]
    block_data["spare"] = data[7:47].decode("ascii").replace("\x00", "")

    return block_data


def _get_metadata_block_08(df, block_info):
    """
    Get metadata block 08.

    Parse the 8th metadata block from an AHI data file and return a
    dictionary containing that metadata.
    """
    # Get the block start and length in bytes
    block_start, block_length = block_info[8]

    # Seek to the start of the block and read the block in its entirety
    df.seek(block_start)
    data = df.read(block_length)

    # Create dictionary for this block
    block_data = {}
    block_data["block_name"] = "navigation_correction_information"
    block_data["block_num"] = np.fromstring(data[0:1], dtype="uint8")[0]
    block_data["block_length"] = np.fromstring(data[1:3], dtype="uint16")[0]
    block_data["center_scan_of_rotation"] = np.fromstring(data[3:7], dtype="float32")[0]
    block_data["center_line_of_rotation"] = np.fromstring(data[7:11], dtype="float32")[
        0
    ]
    block_data["rotation_correction"] = np.fromstring(data[11:19], dtype="float64")[0]
    block_data["num_correction_info"] = np.fromstring(data[19:21], dtype="uint16")[0]

    start = 21
    block_data["line_num_after_rotation"] = np.empty(block_data["num_correction_info"])
    block_data["scan_shift_amount"] = np.empty(block_data["num_correction_info"])
    block_data["line_shift_amount"] = np.empty(block_data["num_correction_info"])
    for info_ind in range(0, block_data["num_correction_info"]):
        block_data["line_num_after_rotation"][info_ind] = np.fromstring(
            data[start : start + 2], dtype="uint16"
        )
        block_data["scan_shift_amount"][info_ind] = np.fromstring(
            data[start + 2 : start + 6], dtype="float32"
        )
        block_data["line_shift_amount"][info_ind] = np.fromstring(
            data[start + 6 : start + 10], dtype="float32"
        )
        start += 10
    block_data["spare"] = data[start : start + 40].decode("ascii").replace("\x00", "")

    return block_data


def _get_metadata_block_09(df, block_info):
    """
    Get metadata block 09.

    Parse the 9th metadata block from an AHI data file and return a
    dictionary containing that metadata.
    """
    # Get the block start and length in bytes
    block_start, block_length = block_info[9]

    # Seek to the start of the block and read the block in its entirety
    df.seek(block_start)
    data = df.read(block_length)

    # Create dictionary for this block
    block_data = {}
    block_data["block_name"] = "observation_time_information"
    block_data["block_num"] = np.fromstring(data[0:1], dtype="uint8")[0]
    block_data["block_length"] = np.fromstring(data[1:3], dtype="uint16")[0]
    block_data["num_ob_times"] = np.fromstring(data[3:5], dtype="uint16")[0]

    start = 5
    block_data["ob_time_line_number"] = np.empty(block_data["num_ob_times"])
    block_data["ob_time"] = np.empty(block_data["num_ob_times"])
    for info_ind in range(0, block_data["num_ob_times"]):
        block_data["ob_time_line_number"][info_ind] = np.fromstring(
            data[start : start + 2], dtype="uint16"
        )
        block_data["ob_time"][info_ind] = np.fromstring(
            data[start + 2 : start + 10], dtype="float64"
        )
        start += 10
    block_data["spare"] = data[start : start + 40].decode("ascii").replace("\x00", "")

    return block_data


def _get_metadata_block_10(df, block_info):
    """
    Get metadata block 10.

    Parse the 10th metadata block from an AHI data file and return a
    dictionary containing that metadata.
    """
    # Get the block start and length in bytes
    block_start, block_length = block_info[10]

    # Seek to the start of the block and read the block in its entirety
    df.seek(block_start)
    data = df.read(block_length)

    # Create dictionary for this block
    block_data = {}
    block_data["block_name"] = "error_information"
    block_data["block_num"] = np.fromstring(data[0:1], dtype="uint8")[0]
    block_data["block_length"] = np.fromstring(data[1:3], dtype="uint16")[0]
    block_data["num_err_info_data"] = np.fromstring(data[3:5], dtype="uint16")[0]

    start = 5
    block_data["err_line_number"] = np.array([])
    block_data["num_err_per_line"] = np.array([])
    for info_ind in range(0, block_data["num_err_info_data"]):
        block_data["err_line_number"].append(
            np.fromstring(data[start : start + 2], dtype="uint16")
        )
        block_data["num_err_per_line"].append(
            np.fromstring(data[start + 2 : start + 4], dtype="uint16")
        )
        start += 4
    block_data["spare"] = data[start : start + 40].decode("ascii").replace("\x00", "")

    return block_data


def _get_metadata_block_11(df, block_info):
    """
    Get metadata block 11.

    Parse the 11th metadata block from an AHI data file and return a
    dictionary containing that metadata.
    """
    # Get the block start and length in bytes
    block_start, block_length = block_info[11]

    # Seek to the start of the block and read the block in its entirety
    df.seek(block_start)
    data = df.read(block_length)

    # Create dictionary for this block
    block_data = {}
    block_data["block_name"] = "spare"
    block_data["block_num"] = np.fromstring(data[0:1], dtype="uint8")[0]
    block_data["block_length"] = np.fromstring(data[1:3], dtype="uint16")[0]
    block_data["spare"] = data[3:259].decode("ascii").replace("\x00", "")

    return block_data


def _get_metadata(df, **kwargs):
    """Gather metadata for the data file and return as a dictionary."""
    metadata = {}
    # Get metadata block info
    try:
        block_info = _get_metadata_block_info(df)
    except IOError as resp:
        raise IOError(resp)

    # Read all 12 blocks
    metadata["block_01"] = _get_metadata_block_01(df, block_info)
    metadata["block_02"] = _get_metadata_block_02(df, block_info)
    metadata["block_03"] = _get_metadata_block_03(df, block_info)
    metadata["block_04"] = _get_metadata_block_04(df, block_info)
    metadata["block_05"] = _get_metadata_block_05(df, block_info)
    metadata["block_06"] = _get_metadata_block_06(df, block_info)
    metadata["block_07"] = _get_metadata_block_07(df, block_info)
    metadata["block_08"] = _get_metadata_block_08(df, block_info)
    metadata["block_09"] = _get_metadata_block_09(df, block_info)
    metadata["block_10"] = _get_metadata_block_10(df, block_info)
    metadata["block_11"] = _get_metadata_block_11(df, block_info)
    # Gather some useful info to the top level
    metadata["path"] = df.name
    # metadata['satellite'] = metadata['block_01']['satellite_name']
    metadata["satellite"] = metadata["block_01"]["satellite_name"]
    # Can this be gotten from the data?
    metadata["sensor"] = "AHI"
    # Make accessable to parent classes that don't know the structure of our metadata
    metadata["num_lines"] = metadata["block_02"]["num_lines"]
    metadata["num_samples"] = metadata["block_02"]["num_samples"]
    return metadata


def _get_files(path):
    """Get a list of file names from the input path."""
    if os.path.isfile(path):
        fnames = [path]
    elif os.path.isdir(path):
        # Temporarily only looking for full disk images.
        # Should change later.
        fnames = glob(path + "/*.DAT")
    else:
        raise IOError("No such file or directory: {0}".format(path))
    return fnames


def _check_file_consistency(metadata):
    """
    Check file consistency.

    Checks to be sure that all input metadata are from the same image time.
    Performs checks on ob_start_time (date only), ob_timeline, ob_area, and
    sub_lon.
    If these are all equal, returns True.
    If any differ, returns False.
    """
    # Checks start dates without comparing times.
    # Times are uncomparable using this field, but are compared below using ob_timeline.
    start_dates = [
        int(metadata[fname]["block_01"]["ob_start_time"]) for fname in metadata.keys()
    ]
    if start_dates[1:] != start_dates[:-1]:
        return False

    # Check the following fields for exact equality.
    #   satellite_name: Must make sure this isn't Himawari-8 and 9 mixed.
    #   ob_timeline: Provides HHMM for each image, so should be the same for all files
    #   from thes same image.
    #   ob_area: Provides the four letter code of the observation area
    #   (e.g. FLDK or JP01).
    #   sub_lon: Just a dummy check to be sure nothing REALLY weird is going on.
    members_to_check = {
        "block_01": {"satellite_name": None, "ob_timeline": None, "ob_area": None},
        "block_03": {"sub_lon": None},
    }

    for block in members_to_check.keys():
        for field in members_to_check[block].keys():
            member_vals = [metadata[fname][block][field] for fname in metadata.keys()]
            # This tests to be sure that all elemnts in member_vals are equal
            # If they aren't all equal, then return False
            if member_vals[1:] != member_vals[:-1]:
                return False
    return True


def sort_by_band_and_seg(metadata):
    """Sort by band and segment."""
    # cfac = metadata['block_03']['CFAC']
    band_number = metadata["block_05"]["band_number"]
    segment_number = metadata["block_07"]["segment_number"]
    # return '{0}_{1:02d}_{2:02d}'.format(cfac, band_number, segment_number)
    return "{0:02d}_{1:02d}".format(band_number, segment_number)


def call(
    fnames,
    metadata_only=False,
    chans=None,
    area_def=None,
    self_register=False,
    test_arg="AHI Default Test Arg",
):
    """
    Read AHI HSD data data from a list of filenames.

    Parameters
    ----------
    fnames : list
        * List of strings, full paths to files
    metadata_only : bool, default=False
        * Return before actually reading data if True
    chans : list of str, default=None
        * List of desired channels (skip unneeded variables as needed).
        * Include all channels if None.
    area_def : pyresample.AreaDefinition, default=None
        * Specify region to read
        * Read all data if None.
    self_register : str or bool, default=False
        * register all data to the specified dataset id (as specified in the
          return dictionary keys).
        * Read multiple resolutions of data if False.

    Returns
    -------
    dict of xarray.Datasets
        * dictionary of xarray.Dataset objects with required Variables and
          Attributes.
        * Dictionary keys can be any descriptive dataset ids.

    See Also
    --------
    :ref:`xarray_standards`
        Additional information regarding required attributes and variables
        for GeoIPS-formatted xarray Datasets.
    """
    return readers.read_data_to_xarray_dict(
        fnames,
        call_single_time,
        metadata_only,
        chans,
        area_def,
        self_register,
    )


def call_single_time(
    fnames,
    metadata_only=False,
    chans=None,
    area_def=None,
    self_register=False,
    test_arg="AHI Default Test Arg",
):
    """
    Read AHI HSD data data from a list of filenames.

    Parameters
    ----------
    fnames : list
        * List of strings, full paths to files
    metadata_only : bool, default=False
        * Return before actually reading data if True
    chans : list of str, default=None
        * List of desired channels (skip unneeded variables as needed).
        * Include all channels if None.
    area_def : pyresample.AreaDefinition, default=None
        * Specify region to read
        * Read all data if None.
    self_register : str or bool, default=False
        * register all data to the specified dataset id (as specified in the
          return dictionary keys).
        * Read multiple resolutions of data if False.

    Returns
    -------
    dict of xarray.Datasets
        * dictionary of xarray.Dataset objects with required Variables and
          Attributes.
        * Dictionary keys can be any descriptive dataset ids.

    See Also
    --------
    :ref:`xarray_standards`
        Additional information regarding required attributes and variables
        for GeoIPS-formatted xarray Datasets.
    """
    process_datetimes = {}
    LOG.debug("AHI reader test_arg: %s", test_arg)
    print_mem_usage("MEMUSG", verbose=False)
    process_datetimes["overall_start"] = datetime.utcnow()
    gvars = {}
    datavars = {}
    adname = "undefined"
    if area_def and self_register:
        raise ValueError("area_def and self_register are mutually exclusive keywords")
    elif area_def:
        adname = area_def.area_id
    elif self_register:
        if self_register not in DATASET_INFO:
            raise ValueError(
                "Unrecognized resolution name requested for self registration: "
                f"{self_register}",
            )
        adname = "FULL_DISK"

    # Get metadata for all input data files
    all_metadata = {}
    for fname in fnames:
        if chans:
            gotone = False
            for chan in chans:
                currchan = chan.replace("BT", "").replace("Ref", "").replace("Rad", "")
                if currchan in fname:
                    gotone = True
            if not gotone:
                LOG.info(
                    "SKIPPING file %s, not needed from channel list %s", fname, chans
                )
                continue
        try:
            all_metadata[fname] = _get_metadata(open(fname, "rb"))
        except IOError as resp:
            LOG.exception("BAD FILE %s skipping", resp)
            if ".bz2" in fname:
                LOG.exception("BAD FILE %s must be bunzip2ed prior to processing", resp)
            continue
        if metadata_only:
            LOG.info("Only need metadata from first file, skipping rest of files")
            break

    # Check to be sure that all input files are from the same image time
    if not _check_file_consistency(all_metadata):
        raise ValueError("Input files inconsistent.")

    # Now put together a dict that shows what we actually got
    # This is largely to help us read in order and only create arrays of the minimum
    # required size. Dict structure is channel{segment{file}}
    # Also build sets containing all segments and channels passed
    file_info = {}
    file_segs = set()
    file_chans = set()

    xarray_obj = xarray.Dataset()
    for md in all_metadata.values():
        ch = "B{0:02d}".format(md["block_05"]["band_number"])
        sn = md["block_07"]["segment_number"]
        if ch not in file_info:
            file_info[ch] = {}
        file_info[ch][sn] = md
        file_segs.add(sn)
        file_chans.add(ch)

    xarray_obj.attrs["file_metadata"] = file_info.copy()

    # Most of the metadata are the same between files.
    # From here on we will just rely on the metadata from a single data file
    # for each resolution.
    res_md = {}
    for res in ["LOW", "MED", "HIGH"]:
        # Find a file file for this resolution: Any one will do
        res_chans = list(set(DATASET_INFO[res]).intersection(file_info.keys()))
        if res_chans:
            # Gets metadata for all available segments for this channel
            segment_info = file_info[res_chans[0]]
            # Get the metadata for any of the segments (doesn't matter which)
            res_md[res] = segment_info[list(segment_info.keys())[0]]

    # If we plan to self register, make sure we requested a resolution that we
    # actually plan to read
    if self_register and self_register not in res_md:
        raise ValueError(
            "Resolution requested for self registration has not been read."
        )

    if len(list(res_md.keys())) == 0:
        raise ValueError(
            "No valid files found in list, make sure .DAT.bz2 are bunzip2-ed: "
            f"{fnames}"
        )

    # Gather metadata
    # Assume the same for all resolutions.   Not true, but close enough.
    highest_md = res_md[list(res_md.keys())[0]]
    start_dt = metadata_to_datetime(
        highest_md, time_var="ob_start_time"
    )  # Assume same for all
    end_dt = metadata_to_datetime(
        highest_md, time_var="ob_end_time"
    )  # Assume same for all
    # mid_dt = start_dt + (end_dt - start_dt)/2
    ob_area = highest_md["block_01"]["ob_area"]
    if ob_area == "FLDK":
        xarray_obj.attrs["sector_name"] = "Full-Disk"
    elif ob_area[0:2] == "JP":
        xarray_obj.attrs["sector_name"] = "Japan-{}".format(ob_area[2:])
    elif ob_area[0] == "R":
        xarray_obj.attrs["sector_name"] = "Regional-{}-{}".format(
            ob_area[1], ob_area[2:]
        )
    else:
        raise ValueError("Unregognized ob_area {}".format(ob_area))
    xarray_obj.attrs["start_datetime"] = start_dt
    xarray_obj.attrs["end_datetime"] = end_dt
    xarray_obj.attrs["source_name"] = "ahi"
    xarray_obj.attrs["data_provider"] = "jma"
    xarray_obj.attrs["platform_name"] = highest_md["block_01"]["satellite_name"].lower()
    xarray_obj.attrs["area_definition"] = area_def
    xarray_obj.attrs["source_file_names"] = [
        os.path.basename(fname) for fname in fnames
    ]

    # If metadata_only requested, return here.
    if metadata_only:
        LOG.info("Only need metadata from first file, returning")
        return {"METADATA": xarray_obj}

    # If one or more channels are missing a segment that other channels have, add it in
    # as "None". This will be set to bad values in the output data
    for ch, segs in file_info.items():
        diff = file_segs.difference(segs.keys())
        if diff:
            segs.update({(sn, None) for sn in diff})

    all_chans_list = []
    for chl in list(ALL_CHANS.values()) + list(ALL_GVARS.values()):
        all_chans_list += chl

    # If specific channels were requested, check them against input data
    # If specific channels were requested, but no files exist for one of the
    # channels, then error
    if chans:
        for chan in chans:
            if chan not in all_chans_list:
                raise ValueError("Requested channel {0} not recognized.".format(chan))
            if chan[0:3] not in file_chans:
                # raise ValueError('Requested channel {0} not found in input data'.
                # format(chan))
                LOG.warning("Requested channel %s not found in input data", chan)

    # If no specific channels were requested, get everything
    if not chans:
        chans = all_chans_list
    # Creates dict whose keys are band numbers in the form B## and whose values are
    # lists containing the data type(s) requested for the band (e.g. Rad, Ref, BT).
    chan_info = {}
    for ch in chans:
        chn = ch[0:3]
        typ = ch[3:]
        if chn not in chan_info:
            chan_info[chn] = []
        chan_info[chn].append(typ)

    # Gather geolocation data
    # Assume datetime the same for all resolutions.  Not true, but close enough.
    # This saves us from having very slightly different solar angles for each channel.
    # Loop over resolutions and get metadata as needed
    # for res in ['HIGH', 'MED', 'LOW']:
    if self_register:
        LOG.info("")
        LOG.info("Getting geolocation information for adname %s.", adname)
        gmd = _get_geolocation_metadata(res_md[self_register])
        fldk_lats, fldk_lons = get_latitude_longitude(gmd, BADVALS, area_def)
        gvars[adname] = get_geolocation(
            start_dt, gmd, fldk_lats, fldk_lons, BADVALS, area_def
        )
        if not gvars[adname]:
            LOG.error(
                "GEOLOCATION FAILED for adname %s DONT_AUTOGEN_GEOLOCATION is: %s",
                adname,
                DONT_AUTOGEN_GEOLOCATION,
            )
            gvars[adname] = {}
    else:
        for res in ["LOW", "MED", "HIGH"]:
            try:
                res_md[res]
            except KeyError:
                continue
            LOG.info("")
            LOG.info(
                "Getting geolocation information for resolution %s for %s", res, adname
            )
            try:
                gmd = _get_geolocation_metadata(res_md[res])
                fldk_lats, fldk_lons = get_latitude_longitude(gmd, BADVALS, area_def)
                gvars[res] = get_geolocation(
                    start_dt, gmd, fldk_lats, fldk_lons, BADVALS, area_def
                )
            except IndexError as resp:
                LOG.exception("SKIPPING apparently no coverage or bad geolocation file")
                raise IndexError(resp)

    LOG.info("Done with geolocation for %s", adname)
    LOG.info("")
    # Read the data
    # Will read all data if area_def is None
    # Will only read required data if an area_def is provided
    for chan, types in chan_info.items():
        if chan not in file_info.keys():
            continue
        LOG.info("Reading %s", chan)
        chan_md = file_info[chan]
        for res, res_chans in DATASET_INFO.items():
            if chan in res_chans:
                break
        if (not self_register) and (res not in gvars.keys() or not gvars[res]):
            LOG.info(
                "We don't have geolocation information for %s for %s skipping %s",
                res,
                adname,
                chan,
            )
            continue
        if not area_def:
            dsname = res
        else:
            dsname = adname

        rad = ref = bt = False
        if "Rad" in types:
            rad = True
        if "Ref" in types:
            ref = True
        if "BT" in types:
            bt = True
        # zoom = 1
        # if self_register:
        #     if self_register == 'HIGH':
        #         if res == 'MED':
        #             zoom = 2
        #         if res == 'LOW':
        #             zoom = 4
        #     elif self_register == 'MED':
        #         if res == 'HIGH':
        #             zoom = 0.5
        #         if res == 'LOW':
        #             zoom = 2
        #     else:
        #         if res == 'HIGH':
        #             zoom = 0.25
        #         if res == 'MID':
        #             zoom = 0.5

        # Need to think about how to implement "zoom" to read only what is needed.
        # This is more complicated than I thought initially.
        # Mostly due to problems with ensuring that zoom produces integer dimensions
        #   and is an integer itself when inverted.
        # data = self.get_data(chan_md, gvars[res], rad, ref, bt, zoom=zoom)
        data = get_data(chan_md, gvars[res], rad, ref, bt)
        for typ, val in data.items():
            if dsname not in datavars:
                datavars[dsname] = {}
            datavars[dsname][chan + typ] = val

    # This needs to be fixed:
    #   remove any unneeded datasets from datavars and gvars
    #   also mask any values below -999.0
    if area_def:
        for res in ["HIGH", "MED", "LOW"]:
            if res not in gvars.keys():
                continue
            if adname not in gvars and "latitude" in gvars[res]:
                gvars[adname] = gvars[res]
            gvars.pop(res)

    if self_register:
        adname = "FULL_DISK"

        # Determine which resolution has geolocation
        LOG.info("Registering to %s", self_register)
        if self_register == "HIGH":
            datavars["FULL_DISK"] = datavars.pop("HIGH")
            for varname, var in datavars["LOW"].items():
                datavars["FULL_DISK"][varname] = zoom(var, 4, order=0)
            datavars.pop("LOW")
            for varname, var in datavars["MED"].items():
                datavars["FULL_DISK"][varname] = zoom(var, 2, order=0)
            datavars.pop("MED")

        elif self_register == "MED":
            datavars["FULL_DISK"] = datavars.pop("MED")
            for varname, var in datavars["LOW"].items():
                datavars["FULL_DISK"][varname] = zoom(var, 2, order=0)
            datavars.pop("LOW")
            for varname, var in datavars["HIGH"].items():
                datavars["FULL_DISK"][varname] = var[::2, ::2]
            datavars.pop("HIGH")

        elif self_register == "LOW":
            datavars["FULL_DISK"] = datavars.pop("LOW")
            for varname, var in datavars["MED"].items():
                datavars["FULL_DISK"][varname] = var[::2, ::2]
            datavars.pop("MED")
            for varname, var in datavars["HIGH"].items():
                datavars["FULL_DISK"][varname] = var[::4, ::4]
            datavars.pop("HIGH")

        else:
            raise ValueError("No geolocation data found.")
    print_mem_usage("MEMUSG", verbose=False)

    # basically just reformat the all_metadata dictionary to
    # reference channel names as opposed to file names..
    band_metadata = get_band_metadata(all_metadata)
    print_mem_usage("MEMUSG", verbose=False)

    # Remove lines and samples arrays.  Not needed.
    for res in gvars.keys():
        try:
            gvars[res].pop("Lines")
            gvars[res].pop("Samples")
            for varname, var in gvars[res].items():
                gvars[res][varname] = np.ma.array(
                    var, mask=gvars[res]["satellite_zenith_angle"].mask
                )
                gvars[res][varname] = np.ma.masked_where(
                    gvars[res]["satellite_zenith_angle"] > 85, gvars[res][varname]
                )
        except KeyError:
            pass
    for ds in datavars.keys():
        if not datavars[ds]:
            datavars.pop(ds)
        else:
            for varname in datavars[ds].keys():
                set_variable_metadata(xarray_obj.attrs, band_metadata, ds, varname)
                datavars[ds][varname] = np.ma.masked_less(datavars[ds][varname], -999.1)
                if "satellite_zenith_angle" in gvars[ds].keys():
                    datavars[ds][varname] = np.ma.masked_where(
                        gvars[ds]["satellite_zenith_angle"] > 85, datavars[ds][varname]
                    )

    print_mem_usage("MEMUSG", verbose=False)
    xarray_objs = {}

    for dsname in datavars.keys():
        xobj = xarray.Dataset()
        xobj.attrs = xarray_obj.attrs.copy()
        for varname in datavars[dsname].keys():
            xobj[varname] = xarray.DataArray(datavars[dsname][varname])
            if re.match(r"B[0-9][0-9]", varname):
                xobj[varname].attrs["channel_number"] = int(varname[1:3])
        for varname in gvars[dsname].keys():
            xobj[varname] = xarray.DataArray(gvars[dsname][varname])
        # if hasattr(xobj, 'area_definition') and xobj.area_definition is not None:
        #     xobj.attrs['interpolation_radius_of_influence'] =
        #     max(xobj.area_definition.pixel_size_x, xobj.area_definition.pixel_size_y)
        # else:
        #     xobj.attrs['interpolation_radius_of_influence'] = 2000
        # Make this a fixed 3000 (1.5 * lowest resolution data) - may want to
        # adjust for different channels.
        xobj.attrs["interpolation_radius_of_influence"] = 3000
        xarray_objs[dsname] = xobj
        # May need to deconflict / combine at some point, but for now just use
        # attributes from any one of the datasets as the METADATA dataset
        xarray_objs["METADATA"] = xobj[[]]
    LOG.info("Done reading AHI data for {}".format(adname))
    LOG.info("")

    print_mem_usage("MEMUSG", verbose=False)
    process_datetimes["overall_end"] = datetime.utcnow()
    from geoips.geoips_utils import output_process_times

    output_process_times(process_datetimes, job_str="AHI HSD Reader")
    return xarray_objs


def set_variable_metadata(xobj_attrs, band_metadata, dsname, varname):
    """
    Set variable metadata.

    MLS 20180914
    Setting xobj_attrs at the variable level for the associated
    channel metadata pulled from the actual netcdf file.
    This will now be accessible from the scifile object.
    Additionally, pull out specifically the band_wavelength and
    attach it to the _varinfo at the variable level - this is
    automatically pulled from the xobj_attrs dictionary
    and set in the variable._varinfo dictionary in scifile/scifile.py
    and scifile/containers.py (see empty_varinfo at the beginning
    of containers.py for dictionary fields that are automatically
    pulled from the appropriate location in the  xobj_attrs
    dictionary and set on the _varinfo dictionary)
    """
    if dsname not in xobj_attrs.keys():
        xobj_attrs[dsname] = {}
    bandname = varname.replace("Rad", "").replace("Ref", "").replace("BT", "")
    if varname not in xobj_attrs[dsname].keys():
        if bandname in band_metadata.keys():
            xobj_attrs[dsname][varname] = {}
            # Store the full metadata dictionary in the scifile metadata
            xobj_attrs[dsname][varname]["all"] = band_metadata[bandname]
            # Set the actual wavelength property on the variable itself
            if (
                "calibration_information" in band_metadata[bandname].keys()
                and "cent_wavelenth"
                in band_metadata[bandname]["calibration_information"].keys()
            ):
                xobj_attrs[dsname][varname]["wavelength"] = band_metadata[bandname][
                    "calibration_information"
                ]["cent_wavelenth"]


def get_band_metadata(all_metadata):
    """
    Get band metadata.

    This method basically just reformats the all_metadata
    dictionary that is set based on the metadata found
    in the netcdf object itself to reference channel
    names as opposed to filenames as the dictionary keys.
    """
    bandmetadata = {}
    for fname in all_metadata.keys():
        bandnum = all_metadata[fname]["block_05"]["band_number"]
        bandmetadata["B%02d" % bandnum] = {}
        for blockname in all_metadata[fname].keys():
            newkey = blockname
            if (
                "block" in blockname
                and "block_name" in all_metadata[fname][blockname].keys()
            ):
                newkey = all_metadata[fname][blockname]["block_name"]
            bandmetadata["B%02d" % bandnum][newkey] = all_metadata[fname][blockname]
    return bandmetadata


def get_data(md, gvars, rad=False, ref=False, bt=False, zoom=1.0):
    """Read data for a full channel's worth of files."""
    # Coordinate arrays for reading
    # Unsure if Lines can ever be None, but the test below was causing an error due to
    # testing the truth value of an entire array.  May need to implement test again here
    # to ensure that gvars['Lines'] is actually something, but testing this method for
    # now.
    # Test against full-disk.  Works for sectors...
    # if ('Lines' in gvars and 'Samples' in gvars) and gvars['Lines'].any()
    # and gvars['Samples'].any():
    if "Lines" in gvars and "Samples" in gvars:
        full_disk = False
        line_inds = gvars["Lines"]
        sample_inds = gvars["Samples"]
        shape = line_inds.shape

        # All required line numbers (used for determining which files to even look at)
        req_lines = set(line_inds.flatten())
    else:
        full_disk = True
        # Assume we are going to make a full-disk image
        smd = md[list(md.keys())[0]]
        nseg = smd["block_07"]["num_segments"]
        lines = smd["num_lines"] * nseg
        samples = smd["num_samples"]
        # sample_inds, line_inds = np.meshgrid(np.arange(samples), np.arange(lines))
        shape = (lines, samples)

        # # Determine dimension sizes
        # lines = []
        # samples = []
        # shell()
        # for seg, smd in md.items():
        #     lines.append(smd['num_lines'])
        #     samples.append(smd['num_samples'])
        # # Sum the nubmer of lines per segment
        # lines = np.sum(lines)
        # # Samples must be the same for all segments
        # samples = set(samples)
        # if len(samples) != 1:
        #     raise ValueError('Number of samples per segment do not match.')
        # samples = list(samples)[0]

        # sample_inds, line_inds = np.meshgrid(np.arange(samples), np.arange(lines))
        # line_inds = line_inds.astype('int32')
        # sample_inds = sample_inds.astype('int32')

        req_lines = set(range(lines))

    # Initialize empty array for channel
    valid_bits = md[list(md.keys())[0]]["block_05"]["valid_bits_per_pixel"]
    # # Ensure zoom produces an integer result
    # zoom_mod = np.mod(np.array(shape), 1/zoom)
    # if np.any(zoom_mod):
    #     raise ValueError('Zoom level does not produce integer dimensions.')
    # counts = np.full(np.int(np.array(shape) * zoom),
    #                  1 + 2**valid_bits, dtype=np.uint16)
    LOG.debug("Making counts array")
    counts = np.full(shape, 1 + 2**valid_bits, dtype=np.uint16)

    # Loop over segments
    for seg, smd in md.items():
        LOG.info("Reading segment {}".format(seg))
        # Skip if we don't have a file for this segment
        if not smd:
            continue

        # Get calibration info
        LOG.debug("Getting calibration info")
        calib = smd["block_05"]
        gain = calib["gain"]
        offset = calib["offset"]
        count_badval = calib["count_badval"]
        count_outbounds = calib["count_outside_scan"]
        valid_bits = smd["block_05"]["valid_bits_per_pixel"]

        # Get info for current file
        LOG.debug("Getting lines and samples")
        path = smd["path"]
        nl = smd["num_lines"]
        ns = smd["num_samples"]
        first_line = smd["block_07"]["segment_first_line"]
        lines = range(first_line, first_line + nl)

        # If required lines and file lines don't intersect, move on to the next file
        if not req_lines.intersection(lines):
            continue

        LOG.debug("Determining indicies in data array.")
        header_len = smd["block_01"]["total_header_length"]
        if not full_disk:
            data_inds = np.where(
                (line_inds >= first_line) & (line_inds < first_line + nl)
            )
            data_lines = line_inds[data_inds] - first_line
            data_samples = sample_inds[data_inds]
            counts[data_inds] = np.memmap(
                path, mode="r", dtype=np.uint16, offset=header_len, shape=(nl, ns)
            )[data_lines, data_samples]
        else:
            first_line -= 1
            last_line = first_line + nl
            counts[first_line:last_line, :] = np.memmap(
                path, mode="r", dtype=np.uint16, offset=header_len, shape=(nl, ns)
            )[:, :]

        LOG.info("Doing the actual read for segment {}".format(seg))

    # It appears that there are values that appear to be good outside the allowable
    # range.  The allowable range is set by the number of valid bits per pixel
    # This number can be 11, 12, or 14
    # These correspond to valid ranges of [0:2048], [0:4096], [0:16384]
    # Here we find all invalid pixels so we can mask later
    outrange_inds = np.where(((counts < 0) | (counts > 2**valid_bits)))
    error_inds = np.where(counts == count_badval)
    offdisk_inds = np.where(counts == count_outbounds)

    # It appears that some AHI data does not correctly set erroneous pixels to
    # count_badval. To fis this, we can find the count value at which radiances become
    # less than or equal to zero. Any counts above that value are bad.
    # Note: This is only for use when calculating brightness temperatures
    # (i.e. when gain is negative)
    if gain < 0:
        root = -offset / gain
        root_inds = np.where(counts > root)
    else:
        root_inds = []

    # Create mask for good values in order to suppress warning from log of negative
    # values.
    # Note: This is only for use when calculating brightness temperatures
    good = np.ones(counts.shape, dtype=bool)
    good[root_inds] = 0
    good[outrange_inds] = 0
    good[error_inds] = 0
    good[offdisk_inds] = 0

    # Create data dict
    data = {}

    # Convert to radiances
    data["Rad"] = counts * gain + offset

    # If reflectance is requested
    # Note the weird memory manipulation to save memory when radiances are not requested
    band_num = calib["band_number"]
    if ref:
        LOG.info("Converting to Reflectance")
        if band_num not in range(1, 7):
            raise ValueError(
                "Unable to calculate reflectances for band #{0}".format(band_num)
            )

        # Get the radiance data
        # Have to do this when using ne
        rad_data = data["Rad"]  # NOQA

        # If we don't need radiances, then reuse the memory
        if not rad:
            data["Ref"] = data.pop("Rad")
        else:
            data["Ref"] = np.empty_like(data["Rad"])

        c_prime = calib["c_prime"]  # NOQA
        ne.evaluate("rad_data * c_prime", out=data["Ref"])

    # If brightness temperature is requested
    # Note the weird memory manipulation to save memory when radiances are not requested
    if bt:
        LOG.info("Converting to Brightness Temperature")
        if band_num not in range(7, 17):
            raise ValueError(
                f"Unable to calculate brightness temperatures for band #{band_num}"
            )

        # Get the radiance data
        # Have to do this when using ne
        rad_data = data["Rad"]  # NOQA

        # If we don't need radiances, then reuse the memory
        if not rad:
            data["BT"] = data.pop("Rad")
        else:
            data["BT"] = np.empty_like(data["Rad"])

        h = calib["planck_const"]
        k = calib["boltz_const"]
        wl = calib["cent_wavelenth"] / 1000000.0  # Initially in microns
        c = calib["speed_of_light"]
        c0 = calib["c0"]  # NOQA
        c1 = calib["c1"]  # NOQA
        c2 = calib["c2"]  # NOQA

        # Make this as in-place as possible
        # Note, rad_data are in units of W/(m**2 sr um)
        log_coeff = (2.0 * h * c**2) / (wl**5)  # NOQA
        dividend = (h * c) / (k * wl)  # NOQA
        ne.evaluate(
            "dividend/log( (log_coeff/(rad_data*1000000.0))+1 )", out=data["BT"]
        )

    for val in data.values():
        LOG.info("Setting badvals")
        val[root_inds] = BADVALS["Root_Test"]
        val[outrange_inds] = BADVALS["Out_Of_Valid_Range"]
        val[offdisk_inds] = BADVALS["Off_Of_Disk"]
        val[error_inds] = BADVALS["Error"]

    return data



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Produce histogram from the given dataset with specified bin size."""
import logging
import numpy

LOG = logging.getLogger(__name__)

interface = "output_formatters"
family = "xarray_data"
name = "histogram_csv"


def call(
    xarray_obj, product_names, output_fnames, min_val=None, max_val=None, num_bins=501
):
    """Produce histogram.

    Parameters
    ----------
    xarray_obj: xarray Dataset

    Returns
    -------
    xarray.Dataset
        Resulting histogram
    """
    import xarray

    prod_xarray = xarray.Dataset()

    from geoips.geoips_utils import copy_standard_metadata
    from geoips.filenames.base_paths import make_dirs
    from os.path import dirname

    copy_standard_metadata(xarray_obj, prod_xarray)
    for product_name in product_names:
        prod_xarray[product_name] = xarray_obj[product_name]

        if product_name == "latitude" or product_name == "longitude":
            continue
        xda = xarray_obj[product_name]
        hist = numpy.histogram(xda, numpy.linspace(min_val, max_val, num_bins + 1))
        prod_xarray[f"{product_name}_histogram"] = xarray.DataArray(
            hist[0], dims=("dim_2")
        )
        prod_xarray["bins"] = xarray.DataArray(hist[1], dims=("dim_3"))
        for ncdf_fname in output_fnames:
            make_dirs(dirname(ncdf_fname))
            with open(ncdf_fname, "w") as fobj:
                prod = prod_xarray[f"{product_name}_histogram"].to_numpy()
                fobj.write(",".join([str(val) for val in list(prod)]))
                LOG.interactive(f"Writing {ncdf_fname}")

    # from geoips.plugins.modules.output_formatters.netcdf_xarray import (
    #     write_xarray_netcdf,
    # )

    # for ncdf_fname in output_fnames:
    #     write_xarray_netcdf(prod_xarray, ncdf_fname)
    return output_fnames



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""MODIS HDF4 reader.

This reader is designed for geoips for importing MODIS data files in hdf4 format
Example files are::

    AQUA:      MYD files
         MYD021KM.A2021004.2005.061.NRT.hdf
         MYD03.A2021004.2005.061.NRT.hdf
         MYD14.A2021004.2005.006.NRT.hdf
    Terra:     MOD files
         MOD021KM.A2021004.2005.061.NRT.hdf
         MOD02HKM.A2021004.2005.061.NRT.hdf
         MOD02QKM.A2021004.2005.061.NRT.hdf
         MOD03.A2021004.2005.061.NRT.hdf
         MOD14.A2021004.2005.006.NRT.hdf

The MOD03 and MOD14 files have the geolocation (lat/lon) and sensor geoometry
infomation, while other files have values at each channels.
"""
# Python Standard Libraries
from datetime import datetime
import logging
from os.path import basename

# Third-Party Libraries
import numpy as np
import xarray as xr

# GeoIPS imports
from geoips.utils.context_managers import import_optional_dependencies

LOG = logging.getLogger(__name__)

with import_optional_dependencies(loglevel="info"):
    """Attempt to import a package and print to LOG.info if the import fails."""
    from pyhdf.HDF import ishdf
    from pyhdf.SD import SD, SDC
    from pyhdf.error import HDF4Error

LOG.info("info on imported functions")

interface = "readers"
family = "standard"
name = "modis_hdf4"
source_names = ["modis"]

# define functions


def parse_metadata(metadatadict):
    """Parse MODIS metadata dictionary."""
    metadata = {}
    for ii in metadatadict.keys():
        # metadata passed by reference - add to it in subroutines.
        if ii == "CoreMetadata.0":
            parse_core_metadata(metadata, metadatadict["CoreMetadata.0"])
        elif ii == "ArchiveMetadata.0":
            parse_archive_metadata(metadata, metadatadict["ArchiveMetadata.0"])
        elif ii == "StructMetadata.0":
            parse_struct_metadata(metadata, metadatadict["StructMetadata.0"])
        else:
            metadata[ii] = metadatadict[ii]
    return metadata


def parse_struct_metadata(metadata, metadatastr):
    """Parse metadata struct."""
    pass


def parse_core_metadata(metadata, metadatastr):
    """Parse core metadata."""
    ii = 0
    lines = metadatastr.split("\n")
    # The HDF4 structure is stored as a big text string in
    # ArchiveMetadata.0, StructMetadata.0, and CoreMetadata.0
    # dictionary entries.  Parsing out the values we need from
    # CoreMetadata.0
    for line in lines:
        # We want to skip END_OBJECT, and only match OBJECT tags. So put a ' '
        # in front of OBJECT Get the value 2 lines past the opening tag
        # (that is the actual value) Remove white space and "s

        # Lines like:
        # '      OBJECT                 = ASSOCIATEDINSTRUMENTSHORTNAME'
        # Skip anything that does not fit that format
        try:
            [typ, field] = [piece.strip() for piece in line.split("=")]
        except ValueError:
            ii += 1
            continue
        for currval in [
            "RANGEENDINGDATE",
            "RANGEBEGINNINGDATE",
            "DAYNIGHTFLAG",
            "SHORTNAME",
        ]:
            if "OBJECT" == typ and currval == field:
                metadata[currval] = lines[ii + 2].split("=")[1].replace('"', "").strip()

        # Also remove the trailing .0000000 from times.
        for currval in ["RANGEBEGINNINGTIME", "RANGEENDINGTIME"]:
            if "OBJECT" == typ and " = " + currval in line:
                metadata[currval] = (
                    lines[ii + 2].split("=")[1].strip().replace('"', "").split(".")[0]
                )

        # These have 'CLASS' in addition to 'NUMVAL' between OBJECT and VALUE.
        # So have to do ii+3
        for currval in ["ASSOCIATEDSENSORSHORTNAME", "ASSOCIATEDPLATFORMSHORTNAME"]:
            if "OBJECT" == typ and currval == field:
                metadata[currval] = lines[ii + 3].split("=")[1].strip().replace('"', "")
        ii += 1


def parse_archive_metadata(metadata, metadatastr):
    """Parse archive metadata."""
    lines = metadatastr.split("\n")
    ii = 0

    # The HDF4 structure is stored as a big text string in
    # ArchiveMetadata.0, StructMetadata.0, and CoreMetadata.0
    # dictionary entries.  Parsing out the values we need from
    # ArchiveMetadata.0
    for line in lines:
        for currattr in [
            "EASTBOUNDINGCOORDINATE",
            "NORTHBOUNDINGCOORDINATE",
            "SOUTHBOUNDINGCOORDINATE",
            "WESTBOUNDINGCOORDINATE",
        ]:
            if " OBJECT" in line and currattr in line:
                # Get the value 2 lines past the opening tag (that is the
                # actual value) Remove white space
                metadata[currattr] = lines[ii + 2].split("=")[1].strip()
        ii += 1


def add_to_xarray(varname, nparr, xobj, cumulative_mask, data_type):
    """Add variable to xarray Dataset."""
    # cumulative_mask is not the best name for this variable. It is used for mask info
    # of field 'varname' so that is not actually a comulative mask. It should be a mask
    # for each variable.
    LOG.info("ADDING %s to xobj", varname)
    if varname not in xobj.variables:
        xobj[varname] = xr.DataArray(nparr)
    else:
        merged_array = np.vstack([xobj[varname].to_masked_array(), nparr])
        xobj[varname] = xr.DataArray(
            merged_array, dims=["dim_" + str(merged_array.shape[0]), "dim_1"]
        )
        cumulative_mask[varname] = xr.DataArray(
            xobj[varname].to_masked_array().mask,
            dims=["dim_" + str(merged_array.shape[0]), "dim_1"],
        )

    # add mask info for 'varname'
    if varname not in list(cumulative_mask.variables.keys()):
        cumulative_mask[varname] = xr.DataArray(xobj[varname].to_masked_array().mask)
    elif cumulative_mask[varname].shape == xobj[varname].shape:
        cumulative_mask[varname] = (
            cumulative_mask[varname] | xobj[varname].to_masked_array().mask
        )


def call(fnames, metadata_only=False, chans=None, area_def=None, self_register=False):
    """Read MODIS Aqua and Terra hdf data files.

    Parameters
    ----------
    fnames : list
        * List of strings, full paths to files
    metadata_only : bool, default=False
        * Return before actually reading data if True
    chans : list of str, default=None
        * List of desired channels (skip unneeded variables as needed).
        * Include all channels if None.
    area_def : pyresample.AreaDefinition, default=None
        * NOT YET IMPLEMENTED
        * Specify region to read
        * Read all data if None.
    self_register : str or bool, default=False
        * NOT YET IMPLEMENTED
        * register all data to the specified dataset id (as specified in the
          return dictionary keys).
        * Read multiple resolutions of data if False.

    Returns
    -------
    dict of xarray.Datasets
        * dictionary of xarray.Dataset objects with required Variables and
          Attributes.
        * Dictionary keys can be any descriptive dataset ids.

    See Also
    --------
    :ref:`xarray_standards`
        Additional information regarding required attributes and variables
        for GeoIPS-formatted xarray Datasets.
    """
    dataset_info = {
        "1KM": [
            "chan20.0Rad",  # emissive 3.750 sfc/cld temp
            "chan21.0Rad",  # emissive 3.750 sfc/cld temp
            "chan22.0Rad",  # emissive 3.959 sfc/cld temp
            "chan23.0Rad",  # emissive 4.050 sfc/cld temp
            "chan24.0Rad",  # emissive 4.465 atm temperature
            "chan25.0Rad",  # emissive 4.515 atm temperature
            "chan27.0Rad",  # emissive 6.715 water vapor
            "chan28.0Rad",  # emissive 7.325 water vapor
            "chan29.0Rad",  # emissive 8.55 sfc/cld temp
            "chan30.0Rad",  # emissive 9.73 ozone
            "chan31.0Rad",  # emissive 11.03 sfc/cld temp
            "chan32.0Rad",  # emissive 12.02 sfc/cld temp
            "chan33.0Rad",  # emissive 13.335 cld top properties
            "chan34.0Rad",  # emissive 13.635 cld top properties
            "chan35.0Rad",  # emissive 13.935 cld top properties
            "chan36.0Rad",  # emissive 14.235 cld top properties
            "chan20.0BT",  # emissive 3.750 sfc/cld temp
            "chan21.0BT",  # emissive 3.750 sfc/cld temp
            "chan22.0BT",  # emissive 3.959 sfc/cld temp
            "chan23.0BT",  # emissive 4.050 sfc/cld temp
            "chan24.0BT",  # emissive 4.465 atm temperature
            "chan25.0BT",  # emissive 4.515 atm temperature
            "chan27.0BT",  # emissive 6.715 water vapor
            "chan28.0BT",  # emissive 7.325 water vapor
            "chan29.0BT",  # emissive 8.55 sfc/cld temp
            "chan30.0BT",  # emissive 9.73 ozone
            "chan31.0BT",  # emissive 11.03 sfc/cld temp
            "chan32.0BT",  # emissive 12.02 sfc/cld temp
            "chan33.0BT",  # emissive 13.335 cld top properties
            "chan34.0BT",  # emissive 13.635 cld top properties
            "chan35.0BT",  # emissive 13.935 cld top properties
            "chan36.0BT",  # emissive 14.235 cld top properties
        ],
        "Fire_Mask": [
            "fire_mask",
        ],
        "HKM": [
            "chan3.0Rad",  # reflective 0.470 land/cld properties
            "chan4.0Rad",  # reflective 0.555 land/cld properties
            "chan5.0Rad",  # reflective 1.24 land/cld properties
            "chan6.0Rad",  # reflective 1.64 land/cld properties
            "chan7.0Rad",  # reflective 2.13 land/cld properties
            "chan16.0Rad",  # reflective 1.64 land/cld properties
            "chan3.0Ref",  # reflective 0.470 land/cld properties
            "chan4.0Ref",  # reflective 0.555 land/cld properties
            "chan5.0Ref",  # reflective 1.24 land/cld properties
            "chan6.0Ref",  # reflective 1.64 land/cld properties
            "chan7.0Ref",  # reflective 2.13 land/cld properties
            "chan16.0Ref",  # reflective 1.64 land/cld properties
        ],
        "QKM": [
            "chan1.0Rad",  # reflective 0.645 land/cld boundaries
            "chan2.0Rad",  # reflective 0.865 land/cld boundaries
            "chan1.0Ref",  # reflective 0.645 land/cld boundaries
            "chan2.0Ref",  # reflective 0.865 land/cld boundaries
        ],
    }
    # @staticmethod

    # from pyhdf.SD import SD, SDC
    # from pyhdf.SD import *

    # fnames=['MYD021KM.A2021004.2005.061.NRT.hdf','MYD03.A2021004.2005.061.NRT.hdf']
    # fnames=['MYD14.A2021004.2005.006.NRT.hdf']
    # fnames=['MYD021KM.A2021004.2005.061.NRT.hdf','MYD03.A2021004.2005.061.NRT.hdf',
    #         'MYD14.A2021004.2005.006.NRT.hdf']

    # process of reading the data
    xarrays = {}
    cumulative_mask = {}  # name of xarray for mask info of all variables
    datapaths = []
    datasettag = ""
    corrections = {}
    corrections_ref = {"aqua": {}, "terra": {}}

    for fname in fnames:  # Loop MODIS files
        LOG.info("Reading file %s", fname)

        # check for a right input MODIS data file
        if ishdf(fname):
            try:
                mf = SD(fname, SDC.READ)
            except HDF4Error:
                LOG.info("wrong input hdf file %s", fname)
                raise

        mf = SD(fname, SDC.READ)
        mf_metadata = parse_metadata(mf.attributes())

        # If start/end datetime happen to vary, adjust here.
        sdt = datetime.strptime(
            mf_metadata["RANGEBEGINNINGDATE"] + mf_metadata["RANGEBEGINNINGTIME"],
            "%Y-%m-%d%H:%M:%S",
        )
        edt = datetime.strptime(
            mf_metadata["RANGEENDINGDATE"] + mf_metadata["RANGEENDINGTIME"],
            "%Y-%m-%d%H:%M:%S",
        )
        cname = mf_metadata["SHORTNAME"]
        xarrays["METADATA"] = xr.Dataset()
        xarrays["METADATA"].attrs["start_datetime"] = sdt
        xarrays["METADATA"].attrs["end_datetime"] = edt
        xarrays["METADATA"].attrs["source_name"] = "modis"
        xarrays["METADATA"].attrs["platform_name"] = mf_metadata[
            "ASSOCIATEDPLATFORMSHORTNAME"
        ].lower()
        xarrays["METADATA"].attrs["data_provider"] = "nasa"
        xarrays["METADATA"].attrs["source_file_names"] = [basename(fname)]
        xarrays["METADATA"].attrs["sample_distance_km"] = 2  # ????
        xarrays["METADATA"].attrs["interpolation_radius_of_influence"] = 3000  # ???
        if metadata_only:
            return xarrays

        if "MOD02QKM" == cname or "MYD02QKM" == cname:
            datapaths = ["EV_250_RefSB"]
            chanlocspaths = ["Band_250M"]
            datasettag = "QKM"
            if datasettag not in cumulative_mask:
                cumulative_mask[datasettag] = xr.Dataset()
                xarrays[datasettag] = xr.Dataset()
                xarrays[datasettag].attrs["source_file_names"] = []
            # corrections_ref[0] = scale
            # corrections_ref[1] = offset
            corrections_ref["aqua"] = {
                "chan1.0": [0.00201941, 0],
                "chan2.0": [0.00327573, 0],
            }
            corrections_ref["terra"] = {
                "chan1.0": [0.00201509, 0],
                "chan2.0": [0.00326201, 0],
            }
        if "MOD02HKM" == cname or "MYD02HKM" == cname:
            datapaths = ["EV_500_RefSB"]
            chanlocspaths = ["Band_500M"]
            datasettag = "HKM"
            if datasettag not in cumulative_mask:
                cumulative_mask[datasettag] = xr.Dataset()
                xarrays[datasettag] = xr.Dataset()
                xarrays[datasettag].attrs["source_file_names"] = []
            # corrections_ref[0] = scale
            # corrections_ref[1] = offset
            corrections_ref["aqua"] = {
                "chan3.0": [0.00155510, 0],
                "chan4.0": [0.00174094, 0],
                "chan5.0": [0.00683729, 0],
                "chan6.0": [0.0134965, 0],
                "chan7.0": [0.0359228, -0],
            }
            corrections_ref["terra"] = {
                "chan3.0": [0.00155013, 0],
                "chan4.0": [0.00173456, 0],
                "chan5.0": [0.00682327, 0],
                "chan6.0": [0.0134728, 0],
                "chan7.0": [0.0358326, -0],
            }
        if "MOD14" == cname or "MYD14" == cname:
            datasettag = "Fire_Mask"
            if datasettag not in cumulative_mask:
                cumulative_mask[datasettag] = xr.Dataset()
                xarrays[datasettag] = xr.Dataset()
                xarrays[datasettag].attrs["source_file_names"] = []
        if "MOD021KM" == cname or "MYD021KM" == cname:
            datapaths = ["EV_1KM_RefSB", "EV_1KM_Emissive"]
            datasettag = "1KM"
            chanlocspaths = ["Band_1KM_RefSB", "Band_1KM_Emissive"]
            if datasettag not in cumulative_mask:
                cumulative_mask[datasettag] = xr.Dataset()
                xarrays[datasettag] = xr.Dataset()
                xarrays[datasettag].attrs["source_file_names"] = []
            # corrections_ref[0] = scale
            # corrections_ref[1] = offset
            corrections_ref["aqua"] = {
                "chan8.0": [0.00185801, 0],
                "chan9.0": [0.00170356, 0],
                "chan10.0": [0.00164244, 0],
                "chan11.0": [0.00172248, 0],
                "chan12.0": [0.00171558, 0],
                "chan13.0": [0.00209848, 0],
                "chan13.5": [0.00209848, -0],
                "chan14.0": [0.00215609, 0],
                "chan14.5": [0.00215609, 0],
                "chan15.0": [0.00250819, 0],
                "chan16.0": [0.00333670, 0],
                "chan17.0": [0.00347492, 0],
                "chan18.0": [0.00372233, 0],
                "chan19.0": [0.00371927, 0],
                "chan26.0": [0.00889511, 0],
            }
            corrections_ref["terra"] = {
                "chan8.0": [0.00185398, 0],
                "chan9.0": [0.00170009, 0],
                "chan10.0": [0.00163386, 0],
                "chan11.0": [0.00171776, 0],
                "chan12.0": [0.00171045, 0],
                "chan13.0": [0.00209057, 0],
                "chan13.5": [0.00209057, -0],
                "chan14.0": [0.00214607, 0],
                "chan14.5": [0.00214607, 0],
                "chan15.0": [0.00249967, 0],
                "chan16.0": [0.00332639, 0],
                "chan17.0": [0.00346232, 0],
                "chan18.0": [0.00370446, 0],
                "chan19.0": [0.00370657, -0],
                "chan26.0": [0.00886857, 0],
            }
            # corrections[chan][0] = (ems1km_um) effective central wavelength in microns
            # corrections[chan][1] = (tcs1km) temperature correction slope
            # corrections[chan][2] = (tci1km) temperature correction intercept
            corrections = {
                "chan20.0": [3.7853, 9.993411e-01, 4.770532e-01],
                "chan21.0": [3.9916, 9.998646e-01, 9.262664e-02],
                "chan22.0": [3.9714, 9.998584e-01, 9.757996e-02],
                "chan23.0": [4.0561, 9.998682e-01, 8.929242e-02],
                "chan24.0": [4.4726, 9.998819e-01, 7.310901e-02],
                "chan25.0": [4.5447, 9.998845e-01, 7.060415e-02],
                "chan27.0": [6.7661, 9.994877e-01, 2.204921e-01],
                "chan28.0": [7.3382, 9.994918e-01, 2.046087e-01],
                "chan29.0": [8.5238, 9.995495e-01, 1.599191e-01],
                "chan30.0": [9.7303, 9.997398e-01, 8.253401e-02],
                "chan31.0": [11.0121, 9.995608e-01, 1.302699e-01],
                "chan32.0": [12.0259, 9.997256e-01, 7.181833e-02],
                "chan33.0": [13.3629, 9.999160e-01, 1.972608e-02],
                "chan34.0": [13.6818, 9.999167e-01, 1.913568e-02],
                "chan35.0": [13.9108, 9.999191e-01, 1.817817e-02],
                "chan36.0": [14.1937, 9.999281e-01, 1.583042e-02],
            }

        # Lat/Lons found in 1km data file appears to be downsampled to 406x271,
        # the lat/lons in MOD03 file are full 1km resolution. hkm and 1km data
        # files appear to have full 1km resolution also (2030x1354)
        # Also, the python resize function appears to just replicate the data,
        # and not interpolate. Which is not what we want.
        if cname == "MOD03" or cname == "MYD03":
            scifile_names = {
                "Latitude": "latitude",
                "Longitude": "longitude",
                "SolarZenith": "solar_zenith_angle",
                "SensorZenith": "satellite_zenith_angle",
                "SolarAzimuth": "solar_azimuth_angle",
                "SensorAzimuth": "satellite_azimuth_angle",
            }

            for datasettag in dataset_info.keys():  # loop the data_type
                #  create a xarray for a data_type if it does not exist
                if datasettag not in cumulative_mask:
                    cumulative_mask[datasettag] = xr.Dataset()
                    xarrays[datasettag] = xr.Dataset()
                    xarrays[datasettag].attrs["source_file_names"] = []

                for (
                    currvar
                ) in scifile_names.keys():  # loop variables assocaited with geo-info
                    sfgvar = scifile_names[currvar]  # name of a field
                    select_data = mf.select(currvar)  # select this field
                    attrs = select_data.attributes()  # get attributes of this field
                    data = select_data.get()  # get the all data of this field

                    # for datasettag in dataset_info.keys():  # loop the data_type
                    # Checking if we need this resolution based on requested
                    # channels
                    if not chans or list(set(chans) & set(dataset_info[datasettag])):
                        LOG.info("    adding " + datasettag + " " + currvar)
                        pass
                    else:
                        continue

                    outdata = data
                    if datasettag == "QKM" or datasettag == "HKM":
                        #  create a xarray for a datasettag if it does not exist
                        # if datasettag not in cumulative_mask:
                        #    cumulative_mask[datasettag] = False
                        #    xarrays[datasettag] = xr.Dataset()

                        factor = 2
                        if datasettag == "QKM":
                            factor = 4
                        outdata = np.zeros(
                            (len(data) * factor, data.shape[1] * factor), data.dtype
                        )
                        x = np.arange(data.shape[0])
                        y = np.arange(data.shape[1])
                        xx = np.linspace(x.min(), x.max(), outdata.shape[0])
                        yy = np.linspace(y.min(), y.max(), outdata.shape[1])
                        from scipy.interpolate import RectBivariateSpline

                        # x, y: 1-D array of coordinate in strickly ascending order
                        # kx, ky (integer,optional): degrees of the bivariate Spline.
                        #                                                   Default is 3
                        # s (float, optional): positive smoothing facter defined for
                        #                     estimation condition. Deault is 0
                        newKernel = RectBivariateSpline(x, y, data, kx=2, ky=2)
                        outdata = newKernel(xx, yy)
                    # Appears lat/lon does not need to be *.01, and
                    # Zenith/Azimuth do. These scale_factors are hard coded
                    # because I can't seem to easily pull them out of the big
                    # text metadata string.
                    factor = 1
                    if "scale_factor" in attrs.keys():
                        factor = attrs["scale_factor"]
                    masked_data = np.ma.masked_equal(
                        np.ma.array(outdata * factor), attrs["_FillValue"]
                    )
                    # variables get propagated to top level in scifile object.
                    # geolocation_variables do not since azimuth/zenith need
                    # to be calculated for each resolution, need to be in
                    # Read lat/lons directly from MOD14.
                    add_to_xarray(
                        sfgvar,
                        masked_data,
                        xarrays[datasettag],
                        cumulative_mask[datasettag],
                        datasettag,
                    )
                    for attrname in attrs:  # add attributes
                        xarrays[datasettag][sfgvar].attrs[attrname] = attrs[attrname]

                # Add attributes
                xarrays[datasettag].attrs["start_datetime"] = sdt
                xarrays[datasettag].attrs["end_datetime"] = edt
                xarrays[datasettag].attrs["source_name"] = "modis"
                xarrays[datasettag].attrs["platform_name"] = mf_metadata[
                    "ASSOCIATEDPLATFORMSHORTNAME"
                ].lower()
                xarrays[datasettag].attrs["data_provider"] = "nasa"
                if (
                    basename(fname)
                    not in xarrays[datasettag].attrs["source_file_names"]
                ):
                    xarrays[datasettag].attrs["source_file_names"] += [basename(fname)]
                xarrays[datasettag].attrs["sample_distance_km"] = 2  # ????
                xarrays[datasettag].attrs[
                    "interpolation_radius_of_influence"
                ] = 3000  # ???

        # elif ('MOD14' == cname or 'MYD14' == cname) and 'fire_mask' in chans:
        elif "MOD14" == cname or "MYD14" == cname:
            # Put shell statement here to figure out how to correct fire_mask
            fire_mask = mf.select("fire mask").get()
            masked_data = np.ma.masked_less(np.ma.array(fire_mask), 7)
            add_to_xarray(
                "fire_mask",
                masked_data,
                xarrays[datasettag],
                cumulative_mask[datasettag],
                datasettag,
            )
            scifile_names = {
                "Latitude": "latitude",
                "Longitude": "longitude",
            }
            # Read lat/lons directly out of MOD14 file, don't need satzenith,etc

            for currvar in scifile_names.keys():
                sfgvar = scifile_names[currvar]
                try:
                    select_data = mf.select(currvar)  # select this field
                except HDF4Error:
                    LOG.warning("SKIPPING %s does not exist in %s", currvar, datasettag)
                    continue
                attrs = select_data.attributes()  # get attributes of this field
                data = select_data.get()  # get the all data of this field
                # data = mf.select(currvar).get()

                add_to_xarray(
                    sfgvar,
                    data,
                    xarrays[datasettag],
                    cumulative_mask[datasettag],
                    datasettag,
                )
                for attrname in attrs:  # add attributes
                    xarrays[datasettag][sfgvar].attrs[attrname] = attrs[attrname]

            # Add attributes
            xarrays[datasettag].attrs["start_datetime"] = sdt
            xarrays[datasettag].attrs["end_datetime"] = edt
            xarrays[datasettag].attrs["source_name"] = "modis"
            xarrays[datasettag].attrs["platform_name"] = mf_metadata[
                "ASSOCIATEDPLATFORMSHORTNAME"
            ].lower()
            xarrays[datasettag].attrs["data_provider"] = "nasa"
            if basename(fname) not in xarrays[datasettag].attrs["source_file_names"]:
                xarrays[datasettag].attrs["source_file_names"] += [basename(fname)]
            xarrays[datasettag].attrs["sample_distance_km"] = 2  # ????
            xarrays[datasettag].attrs["interpolation_radius_of_influence"] = 3000  # ???
        else:
            for ii in range(len(datapaths)):
                datapath = datapaths[ii]
                chanlocspath = chanlocspaths[ii]
                select_alldata = mf.select(datapath)
                chanlocs = mf.select(chanlocspath).get()

                alldata = select_alldata.get()
                attrs = select_alldata.attributes()

                for jj in range(len(chanlocs)):
                    channame = "chan" + str(chanlocs[jj])
                    if (
                        not chans
                        or channame + "Rad" in chans
                        or channame + "Ref" in chans
                        or channame + "BT" in chans
                    ):
                        pass
                    else:
                        continue

                    ind = jj
                    if len(attrs["radiance_offsets"]) == 1:
                        ind = 0
                    data = (alldata[jj] - attrs["radiance_offsets"][ind]) * attrs[
                        "radiance_scales"
                    ][ind]
                    masked_data = np.ma.masked_equal(
                        np.ma.array(data), attrs["_FillValue"]
                    )
                    masked_data = np.ma.masked_greater(
                        masked_data, attrs["valid_range"][1]
                    )
                    masked_data = np.ma.masked_less(
                        masked_data, attrs["valid_range"][0]
                    )

                    LOG.info(
                        "    Adding channame: "
                        + str(channame + "Rad")
                        + " offset: "
                        + str(attrs["radiance_offsets"][ind])
                        + " scale: "
                        + str(attrs["radiance_scales"][ind])
                    )
                    add_to_xarray(
                        channame + "Rad",
                        masked_data,
                        xarrays[datasettag],
                        cumulative_mask[datasettag],
                        datasettag,
                    )
                    for attrname in attrs:  # add attributes
                        xarrays[datasettag][channame + "Rad"].attrs[attrname] = attrs[
                            attrname
                        ]

                    if channame in corrections.keys():
                        cor = corrections[channame]
                        h = 6.626176e-34
                        c = 2.99792458e8
                        bc = 1.380658e-23

                        data1 = np.log(
                            1
                            + 2
                            * h
                            * c
                            * c
                            / (
                                cor[0]
                                * cor[0]
                                * cor[0]
                                * cor[0]
                                * cor[0]
                                * 1.0e-30
                                * masked_data
                                * 1.0e6
                            )
                        )
                        data2 = h * c / (bc * cor[0] * 1.0e-6) / data1
                        # masked_data = ((data2-cor[2])/cor[1]) - 273.15
                        # By default, GeoIPS readers should store BT data internally in
                        # Kelvin.
                        masked_data = (data2 - cor[2]) / cor[1]

                        LOG.info(
                            "    Adding channame: "
                            + str(channame + "BT")
                            + " offset: "
                            + str(attrs["radiance_offsets"][ind])
                            + " scale: "
                            + str(attrs["radiance_scales"][ind])
                            + " ems1km: "
                            + str(cor[0])
                            + " tcs1km: "
                            + str(cor[1])
                            + " tci1km: "
                            + str(cor[2])
                            + " min: "
                            + str(masked_data.min())
                            + " max: "
                            + str(masked_data.max())
                        )
                        add_to_xarray(
                            channame + "BT",
                            masked_data,
                            xarrays[datasettag],
                            cumulative_mask[datasettag],
                            datasettag,
                        )
                        xarrays[datasettag][channame + "BT"].attrs["units"] = "Kelvin"
                    if (
                        channame
                        in corrections_ref[
                            mf_metadata["ASSOCIATEDPLATFORMSHORTNAME"].lower()
                        ].keys()
                    ):
                        corrected_data = np.ma.masked_equal(
                            np.ma.array(data), attrs["_FillValue"]
                        )
                        corrected_data = np.ma.masked_greater(
                            masked_data, attrs["valid_range"][1]
                        )
                        corrected_data = np.ma.masked_less(
                            masked_data, attrs["valid_range"][0]
                        )
                        offset = corrections_ref[
                            mf_metadata["ASSOCIATEDPLATFORMSHORTNAME"].lower()
                        ][channame][1]
                        scale = corrections_ref[
                            mf_metadata["ASSOCIATEDPLATFORMSHORTNAME"].lower()
                        ][channame][0]
                        corrected_data += offset
                        corrected_data *= 100.0 * scale
                        LOG.info(
                            "    Adding channame: "
                            + str(channame + "Ref")
                            + " offset: "
                            + str(offset)
                            + " scale: "
                            + str(scale)
                        )
                        add_to_xarray(
                            channame + "Ref",
                            corrected_data,
                            xarrays[datasettag],
                            cumulative_mask[datasettag],
                            datasettag,
                        )

            # Add attributes
            xarrays[datasettag].attrs["start_datetime"] = sdt
            xarrays[datasettag].attrs["end_datetime"] = edt
            xarrays[datasettag].attrs["source_name"] = "modis"
            xarrays[datasettag].attrs["platform_name"] = mf_metadata[
                "ASSOCIATEDPLATFORMSHORTNAME"
            ].lower()
            xarrays[datasettag].attrs["data_provider"] = "nasa"
            if basename(fname) not in xarrays[datasettag].attrs["source_file_names"]:
                xarrays[datasettag].attrs["source_file_names"] += [basename(fname)]
            xarrays[datasettag].attrs["sample_distance_km"] = 2  # ????
            xarrays[datasettag].attrs["interpolation_radius_of_influence"] = 3000  # ???

    # compbine output fields
    xarray_returns = {}
    for dtype in xarrays.keys():
        LOG.info("Trying xarrays[%s]", dtype)
        if dtype != "METATDATA" and "latitude" not in list(
            xarrays[dtype].variables.keys()
        ):
            LOG.info("No data read for dataset %s, removing from xarray list", dtype)
            continue
        if dtype != "METADATA":
            for varname in list(xarrays[dtype].variables.keys()):
                xarrays[dtype][varname] = xarrays[dtype][varname].where(
                    ~cumulative_mask[dtype][varname]
                )

        LOG.info("Adding xarrays[%s]", dtype)
        xarray_returns[dtype] = xarrays[dtype]

    if len(list(xarray_returns.values())) == 0:
        raise IOError("No data found for requested channels %s", chans)

    xarray_returns["METADATA"] = list(xarray_returns.values())[0][[]]

    return xarray_returns



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Databases interface module."""

from geoips.interfaces.base import BaseModuleInterface


class DatabasesInterface(BaseModuleInterface):
    """Interface for database table writers/quieriers."""

    name = "databases"
    required_args = {
        "xarray_area_def_to_table": ["product_filename", "xarray_obj", "area_def"],
        "xarray_dict_to_table": ["product_filename", "xarray_dict"],
        "memusg_stats_to_table": [],
        "overpass_dict_to_postgres_table": [
            "schema_name",
            "table_name",
            "overpass_dict",
        ],
        "overpass_dict_to_sqlite_table": ["overpass_dict", "dbname"],
        "schema_table_sector_opasses_to_postgres_table": [
            "schema_name",
            "table_name",
            "sector_name",
            "sat_overpasses",
        ],
        "schema_table_storm_opasses_to_postgres_table": [
            "schema_name",
            "table_name",
            "storm_id",
            "storm_time",
            "satellite_overpasses",
        ],
        "storm_opasses_db_to_postgres_table": [
            "storm_id",
            "storm_time",
            "satellite_overpasses",
            "dbname",
        ],
        "query_overpass_table": ["window_start", "window_end"],
        "query_product_table": ["window_start", "window_end"],
    }
    required_kwargs = {
        "xarray_area_def_to_table": [],
        "xarray_dict_to_table": [],
        "memusg_stats_to_table": [],
        "overpass_dict_to_postgres_table": [],
        "overpass_dict_to_sqlite_table": [],
        "storm_opasses_db_to_postgres_table": [],
        "schema_table_storm_opasses_to_postgres_table": [],
        "schema_table_sector_opasses_to_postgres_table": [],
        "query_overpass_table": [],
        "query_product_table": [],
    }


databases = DatabasesInterface()



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Unit test for GeoIPS CLI `list unit-tests` command.

See geoips/commandline/ancillary_info/cmd_instructions.yaml for more information.
"""

from glob import glob
from importlib.resources import files
from os import listdir
from os.path import basename
import pytest
import logging

from tests.unit_tests.commandline.cli_top_level_tester import BaseCliTest

LOG = logging.getLogger(__name__)


class TestGeoipsListUnitTests(BaseCliTest):
    """Unit Testing Class for List Unit Tests Command."""

    @property
    def command_combinations(self):
        """A list of every possible call signature for the GeoipsListUnitTests command.

        This includes failing cases as well.
        """
        if not hasattr(self, "_cmd_list"):
            base_args = ["geoips", "list", "unit-tests"]
            alias_args = ["geoips", "ls", "unit-tests"]
            self._cmd_list = [base_args, alias_args]
            for argset in [base_args, alias_args]:
                for pkg_name in self.plugin_package_names:
                    self._cmd_list.append(argset + ["-p", pkg_name])
            # Add argument list which invokes the help message for this command
            self._cmd_list.append(["geoips", "list", "unit-tests", "-h"])
            self._cmd_list.append(["geoips", "ls", "unit-tests", "-h"])
            # Add argument list with a non-existent package
            self._cmd_list.append(
                ["geoips", "list", "unit-tests", "-p", "non_existent_package"]
            )
            self._cmd_list.append(
                ["geoips", "ls", "unit-tests", "-p", "non_existent_package"]
            )
        return self._cmd_list

    def check_error(self, args, error):
        """Ensure that the 'geoips list unit-tests ...' error output is correct.

        Parameters
        ----------
        args: 2D list of str
            - The arguments used to call the CLI (expected to fail)
        error: str
            - Multiline str representing the error output of the CLI call
        """
        # If there is no test directory, xfail.
        if "unit-tests: error: No unit test directory found under " in error:
            LOG.warning("No unit test directory found, please add unit tests")
            pytest.xfail(str(error))
        editable = self.assert_non_editable_error_or_wrong_package(args, error)
        if editable:
            # bad command has been provided, check the contents of the error message
            assert args != ["geoips", "list", "unit-tests"] and args != [
                "geoips",
                "ls",
                "unit-tests",
            ]
            assert "usage: To use, type `geoips list unit-tests" in error

    def check_output(self, args, output):
        """Ensure that the 'geoips list unit-tests ...' successful output is correct.

        Parameters
        ----------
        args: 2D list of str
            - The arguments used to call the CLI
        output: str
            - Multiline str representing the output of the CLI call
        """
        if "usage: To use, type" in output:
            # -h has been called, check help message contents for this command
            assert "-h" in args
            assert "usage: To use, type `geoips list unit-tests" in output
        # If there is no test directory, xfail.
        elif "unit-tests: error: No unit test directory found under " in output:
            LOG.warning("No unit test directory found, please add unit tests")
            pytest.xfail(str(output))
        else:
            # The args provided are valid, so test that the output is actually correct
            # Assert that the correct headers exist in the CLI output
            headers = ["GeoIPS Package", "Unit Test Directory", "Unit Test Name"]
            for header in headers:
                assert header in output
            if "-p" in args:
                pkg_name = args[-1]
            else:
                pkg_name = "geoips"
            unit_test_dir = str(files(pkg_name) / "../tests/unit_tests")
            # Assert that we found every unit test
            for subdir_name in listdir(unit_test_dir):
                for unit_test in sorted(
                    glob(f"{unit_test_dir}/{subdir_name}/test_*.py")
                ):
                    assert basename(unit_test) in output


test_sub_cmd = TestGeoipsListUnitTests()


@pytest.mark.parametrize(
    "args",
    test_sub_cmd.command_combinations,
    ids=test_sub_cmd.generate_id,
)
def test_command_combinations(monkeypatch, args):
    """Test all 'geoips list unit-tests ...' commands.

    This test covers every valid combination of commands for the
    'geoips list unit-tests' command. We also test invalid commands, to ensure that
    the proper help documentation is provided for those using the command incorrectly.

    Parameters
    ----------
    args: 2D array of str
        - List of arguments to call the CLI with (ie. ['geoips', 'list', 'unit-tests'])
    """
    test_sub_cmd.test_command_combinations(monkeypatch, args)



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Standard geoips filename production."""

# Python Standard Libraries
import logging

from os.path import join as pathjoin

from geoips.filenames.base_paths import PATHS as gpaths

LOG = logging.getLogger(__name__)

interface = "filename_formatters"
family = "standard"
name = "geoips_fname"


def call(
    area_def,
    xarray_obj,
    product_name,
    coverage=None,
    output_type="png",
    output_type_dir=None,
    product_dir=None,
    product_subdir=None,
    source_dir=None,
    basedir=gpaths["ANNOTATED_IMAGERY_PATH"],
    output_dict=None,
):
    """Create GeoIPS standard filenames, sector-based subdirs.

    This uses the sector specification (continent, country, area, subarea,
    state, city), product name, source name, and platform name
    to generate a full unique path, as well as additional attributes to
    create a fully unique file name.
    """
    # from geoips.xarray_utils.time import get_min_from_xarray_time
    # start_dt = get_min_from_xarray_time(xarray_obj, 'time')
    start_dt = xarray_obj.start_datetime

    resolution = max(area_def.pixel_size_x, area_def.pixel_size_y) / 1000.0
    # This means area_extent is defined in units of degrees longitude, not meters.
    if hasattr(area_def, "area_extent"):
        if area_def.area_extent == area_def.area_extent_ll:
            # 111km per degree longitude.
            resolution = max(area_def.resolution[0], area_def.resolution[1]) * 111.0

    kwargs = {}
    sector_info = (
        area_def.sector_info
        if "region" not in area_def.sector_info.keys()
        else area_def.sector_info["region"]
    )
    if "continent" in sector_info:
        kwargs["continent"] = sector_info["continent"]
    if "country" in sector_info:
        kwargs["country"] = sector_info["country"]
    if "area" in sector_info:
        kwargs["area"] = sector_info["area"]
    if "subarea" in sector_info:
        kwargs["subarea"] = sector_info["subarea"]
    if "state" in sector_info:
        kwargs["state"] = sector_info["state"]
    if "city" in sector_info:
        kwargs["city"] = sector_info["city"]

    extra = "{0:0.1f}".format(resolution).replace(".", "p")
    web_fname = assemble_geoips_fname(
        basedir=basedir,
        product_name=product_name,
        source_name=xarray_obj.source_name,
        platform_name=xarray_obj.platform_name,
        sector_name=area_def.area_id,
        coverage=coverage,
        resolution=resolution,
        product_datetime=start_dt,
        output_type=output_type,
        data_provider=xarray_obj.data_provider,
        extra=extra,
        product_dir=product_dir,
        source_dir=source_dir,
        **kwargs,
    )
    return web_fname


def assemble_geoips_fname(
    basedir,
    product_name,
    source_name,
    platform_name,
    sector_name,
    coverage,
    resolution,
    product_datetime,
    output_type="png",
    data_provider=None,
    extra=None,
    product_dir=None,
    source_dir=None,
    continent=None,
    country=None,
    area=None,
    subarea=None,
    state=None,
    city=None,
):
    """Produce full output product path from product / sensor specifications.

    standard web paths are of the format:
        <basedir>/<continent>-<country>-<area>/<subarea>-<state>-<city>/
        <productname>/<sensorname>``
    standard filenames are of the format:
        <date{%Y%m%d}>.<time{%H%M%S}>.<satname>.<sensorname>.<productname>.
        <sectorname>.<coverage>.<dataprovider>.<extra>

    Parameters
    ----------
    basedir : str
        Full path to base directory of final product.
    product_name : str
        Name of product
    source_name : str
        Name of data source (sensor)
    platform_name : str
        Name of platform (satellite)
    coverage : float
        Image coverage, float between 0.0 and 100.0
    resolution : float
        Image resolution, float greater than 0.0
    product_datetime : datetime.datetime
        Datetime object - start time of data used to generate product.

    Other Parameters
    ----------------
    output_type : str, optional
        file extension type, default is png
    data_provider :  str, optional
        String to include in filename "data_provider" field
    extra : str, optional
        String to include in filename "extra" field, default is None
        If None, use fillval of 'x'
    continent : str, optional
        String to include in filename "continent" field, default is None
        If None, use fillval of 'x'
    country : str, optional
        String to include in filename "country" field, default is None
        If None, use fillval of 'x'
    area : str, optional
        String to include in filename "area" field, default is None
        If None, use fillval of 'x'
    subarea : str, optional
        String to include in filename "subarea" field, default is None
        If None, use fillval of 'x'
    state : str, optional
        String to include in filename "state" field, default is None
        If None, use fillval of 'x'
    city : str, optional
        String to include in filename "city" field, default is None
        If None, use fillval of 'x'
    """
    fillval = "x"
    if continent is None:
        continent = fillval
    if country is None:
        country = fillval
    if area is None:
        area = fillval
    if subarea is None:
        subarea = fillval
    if state is None:
        state = fillval
    if city is None:
        city = fillval
    if data_provider is None:
        data_provider = fillval
    if extra is None:
        extra = fillval
    if product_dir is None:
        product_dir = product_name
    if source_dir is None:
        source_dir = source_name
    LOG.info(basedir)
    path = pathjoin(
        basedir,
        "{0}-{1}-{2}".format(continent, country, area),
        "{0}-{1}-{2}".format(subarea, state, city),
        product_dir,
        source_dir,
    )
    # source_dir,
    # '{0:0.1f}'.format(resolution).replace('.', 'p'))
    # fname = '<date{%Y%m%d}>.<time{%H%M%S}>.<satname>.<sensorname>
    #          .<productname>.<sectorname>
    #          .<coverage>.<dataprovider>.<extra>'
    fname = ".".join(
        [
            product_datetime.strftime("%Y%m%d"),
            product_datetime.strftime("%H%M%S"),
            platform_name,
            source_name,
            product_name,
            sector_name,
            "{0:0.2f}".format(coverage).replace(".", "p"),
            data_provider,
            str(extra),
        ]
    )
    fname = "{0}.{1}".format(fname, output_type)
    return pathjoin(path, fname)


def geoips_fname_remove_duplicates(fname, mins_to_remove=10, remove_files=False):
    """remove_duplicates function currently not defined.

    If defined, this function will identify duplicate files, and remove
    appropriately.  It should identify duplicates specifically based
    on the format/location of the geoips_fname formatted files.

    Currently returns empty lists.  When defined, will return a list of
    deleted files and a list of saved files.
    """
    LOG.info("MUST ADD LOGIC TO REMOVE STANDARD GEOIPS FILENAME DUPLICATES")
    return [], []
